import copy
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


def convert_raw_stage_to_present(
    hub,
    raw_resource: Dict[str, Any],
    resource_id: str,
) -> Dict[str, Any]:
    resource_parameters = OrderedDict(
        {
            "stageName": "name",
            "deploymentId": "deployment_id",
            "cacheClusterEnabled": "cache_cluster_enabled",
            "cacheClusterSize": "cache_cluster_size",
            "variables": "variables",
            "description": "description",
            "canarySettings": "canary_settings",
            "tracingEnabled": "tracing_enabled",
            "documentationVersion": "documentation_version",
            "tags": "tags",
        }
    )

    resource_translated = {
        "resource_id": resource_id,
        "rest_api_id": resource_id.split("-")[0],
        "name": resource_id.split("-")[1],
    }

    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw) is not None:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated


async def update(
    hub,
    ctx,
    old_state: Dict[str, Any],
    updateable_parameters: Dict[str, Any],
):
    r"""

    Updates an API Stage Resource.

    Args:
        old_state(dict):
            Previous state of the resource.

        updateable_parameters(dict):
            Parameters from SLS File.

    Returns:
        {"result": True|False, "comment": A message List, "ret": None|Dict}

    """

    result = dict(comment=(), result=True, ret=None)
    patch_ops = []

    new_state = copy.deepcopy(old_state)
    updated = False
    key_final = ""

    for key, value in updateable_parameters.items():
        if key == "cache_cluster_enabled":
            key_final = "cacheClusterEnabled"
        elif key == "cache_cluster_size":
            key_final = "cacheClusterSize"
        elif key == "deployment_id":
            key_final = "deploymentId"
        elif key == "description":
            key_final = "description"
        elif key == "variables":
            key_final = "variables"
        elif key == "tracing_enabled":
            key_final = "tracingEnabled"

        if value != old_state.get(key):
            patch_ops.append(
                {
                    "op": "replace",
                    "path": f"/{key_final}",
                    "value": str(value)
                    if not isinstance(value, List)
                    else ",".join(value),
                }
            )
            new_state[key] = value
            updated = True
        elif old_state.get(key) is None and value is not None:
            patch_ops.append(
                {
                    "op": "add",
                    "path": f"/{key_final}",
                    "value": str(value)
                    if not isinstance(value, List)
                    else ",".join(value),
                }
            )
            new_state[key] = value
            updated = True

    if ctx.get("test", False):
        if updated:
            result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.apigateway.stage",
                name=old_state["name"],
            )
            result["ret"] = new_state
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.apigateway.stage",
                name=old_state["name"],
            )
        return result

    if updated:
        update_ret = await hub.exec.boto3.client.apigateway.update_stage(
            ctx,
            restApiId=old_state["rest_api_id"],
            stageName=old_state["name"],
            patchOperations=patch_ops,
        )
        if not update_ret["result"]:
            result["result"] = False
            result["comment"] = update_ret["comment"]
            return result
        else:
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.apigateway.stage",
                name=old_state["name"],
            )
            result["ret"] = new_state
    else:
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.apigateway.stage", name=old_state["name"]
        )

    return result



"""Exec module for managing Amazon API Gateway Deployments."""


async def get(
    hub,
    ctx,
    resource_id: str,
    rest_api_id: str,
    name: str = None,
):
    """Get an API Gateway Deployments from AWS. If more than one resource is found, the first resource returned from AWS
    will be used. The function returns None when no resource is found.

    Args:
        resource_id(str):
            Resource ID of the AWS Deployment.

        rest_api_id(str):
            The string identifier of the associated RestApi.

        name(str, Optional):
            An Idem name of the API Gateway Deployment.

    Returns:
        .. code-block:: python

            {"result": True|False, "comment": A message List, "ret": Dict[str, Any]}

     Examples:

        Calling this exec module function from the cli with resource_id

        .. code-block:: bash

            idem exec aws.apigateway.deployment.get resource_id="resource_id" rest_api_id="rest_api_id"
            name="umanaged_deployment"


        Calling this exec module function from within a state module in pure python

        .. code-block:: yaml

            my_unmanaged_deployment:
              exec.run:
                - path: aws.apigateway.deployment.get
                - kwargs:
                    rest_api_id: 5hkdu2
                    resource_id: f8kils
                    name: test-deployments-name


    """

    result = dict(comment=[], ret=None, result=True)
    ret = await hub.exec.boto3.client.apigateway.get_deployment(
        ctx, restApiId=rest_api_id, deploymentId=resource_id
    )

    if not ret["result"]:
        if "NotFoundException" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.apigateway.deployment", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    if not ret["ret"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.apigateway.deployment", name=name
            )
        )
        return result

    else:
        result[
            "ret"
        ] = hub.tool.aws.apigateway.deployment.convert_raw_deployment_to_present(
            raw_resource=ret["ret"],
            resource_id=resource_id,
            rest_api_id=rest_api_id,
            idem_resource_name=name,
        )

    return result



"""Build CRUD function definitions for a resource to be used in resource plugins"""
from typing import Any
from typing import Dict


def parse_functions(
    hub,
    session: "boto3.session.Session",
    aws_service_name: str,
    resource_name: str,
    functions: dict,
) -> Dict[str, Any]:
    resource_methods = {}
    try:
        client = session.client(service_name=aws_service_name, region_name="us-west-2")

        resource_methods["get"] = hub.pop_create.aws.function.parse(
            client=client,
            aws_service_name=aws_service_name,
            resource_name=resource_name,
            func_name=functions.get("get"),
        )

        resource_methods["list"] = hub.pop_create.aws.function.parse(
            client=client,
            aws_service_name=aws_service_name,
            resource_name=resource_name,
            func_name=functions.get("list"),
        )

        resource_methods["create"] = hub.pop_create.aws.function.parse(
            client=client,
            aws_service_name=aws_service_name,
            resource_name=resource_name,
            func_name=functions.get("create"),
        )

        resource_methods["delete"] = hub.pop_create.aws.function.parse(
            client=client,
            aws_service_name=aws_service_name,
            resource_name=resource_name,
            func_name=functions.get("delete"),
        )

        resource_methods["update"] = hub.pop_create.aws.function.parse(
            client=client,
            aws_service_name=aws_service_name,
            resource_name=resource_name,
            func_name=functions.get("update"),
        )

        for func_name in functions.get("tools", []):
            try:
                resource_methods[func_name] = hub.pop_create.aws.function.parse(
                    client=client,
                    aws_service_name=aws_service_name,
                    resource_name=resource_name,
                    func_name=func_name,
                )
            except Exception as err:
                hub.log.error(
                    f"Error when generating resource's action definitions for {resource_name}:{func_name}: {err.__class__.__name__}: {err}"
                )

    except Exception as err:
        hub.log.error(
            f"Error when generating resource's action definitions for {resource_name}: {err.__class__.__name__}: {err}"
        )

    return resource_methods



"""Utility functions for Rds Db Proxys."""
from typing import Any
from typing import Dict


async def convert_raw_resource_to_present_async(
    hub, ctx, idem_resource_name: str, resource_id: str, raw_resource: dict
) -> Dict[str, Any]:
    r"""
    Convert raw resource of db_proxy type into present format.

    """

    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}

    resource_parameters = {
        "DBProxyName": "db_proxy_name",
        "EngineFamily": "engine_family",
        "VpcSecurityGroupIds": "vpc_security_group_ids",
        "VpcSubnetIds": "vpc_subnet_ids",
        "Auth": "auth",
        "RoleArn": "role_arn",
        "RequireTLS": "require_tls",
        "IdleClientTimeout": "idle_client_timeout",
        "DebugLogging": "debug_logging",
    }

    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource and raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    # Get it from raw_resource or explicitly retrieved tags
    resource_tags_list = raw_resource.get("Tags") or raw_resource.get("TagList")

    if resource_tags_list:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            resource_tags_list
        )
    else:
        tags_ret = await hub.tool.aws.rds.tag.get_tags_for_resource(
            ctx, resource_arn=raw_resource.get("DBProxyArn")
        )
        if tags_ret["result"]:
            resource_translated["tags"] = tags_ret.get("ret", {})

    return resource_translated



import copy
from collections import OrderedDict
from typing import Any
from typing import Dict


def convert_raw_resource_to_present(
    hub,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
    rest_api_id: str = None,
) -> Dict[str, Any]:
    """Convert AWS API Gateway Resource to a common idem present state.

    Args:
        raw_resource (dict[str, Any]): The AWS response to convert.

        idem_resource_name (str, Optional): The resource name if applicable,
            otherwise resource name will be taken from raw_resource.

        rest_api_id (str, Optional): Rest Api id associated with the Rest Api to which the Resource belongs.

    Returns:
        dict[str, Any]: Common idem present state
    """

    resource_id = raw_resource.get("id")

    if idem_resource_name:
        name = idem_resource_name
    else:
        name = raw_resource.get("pathPart")

    resource_parameters = OrderedDict(
        {
            "parentId": "parent_id",
            "pathPart": "path_part",
            "path": "path",
        }
    )

    resource_translated = {
        "name": name,
        "resource_id": resource_id,
        "rest_api_id": rest_api_id,
    }
    for parameter_raw, parameter_present in resource_parameters.items():
        resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated


async def update_resource(
    hub, ctx, old_state: Dict[str, Any], update_parameters: Dict[str, any]
):
    """Updates an API Gateway Resource

    Args:
        old_state (dict): Previous state o  f resource

        update_parameters (dict): Parameters from SLS File

    Returns:
        dict[str, Any]
    """

    result = dict(comment=(), result=True, ret=None)
    patch_ops = []

    new_state = copy.deepcopy(old_state)
    updated = False

    for key, value in update_parameters.items():
        if key == "parent_id":
            key_final = "parentId"
        else:
            key_final = "pathPart"

        if old_state.get(key) and value != old_state.get(key):
            patch_ops.append(
                {
                    "op": "replace",
                    "path": f"/{key_final}",
                    "value": str(value)
                    if not isinstance(value, list)
                    else ",".join(value),
                }
            )
            updated = True

        elif old_state.get(key) is None:
            patch_ops.append(
                {
                    "op": "add",
                    "path": f"/{key_final}",
                    "value": str(value)
                    if not isinstance(value, list)
                    else ",".join(value),
                }
            )

            new_state[key] = value
            updated = True

    if ctx.get("test", False):
        if updated:
            result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.apigateway.resource", name=old_state["name"]
            )
            result["ret"] = new_state
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.apigateway.resource", name=old_state["name"]
            )
        return result

    if updated:
        update_ret = await hub.exec.boto3.client.apigateway.update_resource(
            ctx,
            restApiId=old_state["rest_api_id"],
            resourceId=old_state["resource_id"],
            patchOperations=patch_ops,
        )
        if not update_ret["result"]:
            result["result"] = False
            result["comment"] = update_ret["comment"]
            return result
        else:
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.apigateway.resource", name=old_state["name"]
            )
            result["ret"] = new_state
    else:
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.apigateway.resource", name=old_state["name"]
        )

    return result



def unescape_ascii(hub, input_str: str):
    """
    If '*' and '@' are present in any attribute values, boto3 API returns the attribute values
    with ASCII equivalent of '*' and '@'. we need to replace ASCII characters with equivalent string characters
    to compare with user input.

    Args:
        input_str: value of the attribute

    Returns:
        cleaned attribute value by replacing the ASCII characters with equivalent string characters

    """
    return input_str.encode("utf-8").decode("unicode_escape")



"""State module for managing Amazon Config Recorder status."""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]

TREQ = {
    "present": {
        "require": [
            "aws.config.config_recorder.present",
            "aws.config.delivery_channel.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    recording: bool = False,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Set a new configuration recorder status to record the selected resource configurations.

    Args:
        name(str):
            The name of the recorder.

        recording (bool, Optional):
            Specifies recording status of the configuration recorder. Default is False.

        resource_id (str, Optional):
            The name of the recorder.

    Request syntax:
        .. code-block:: sls

            [aws-config-recorder-status]:
              aws.config.config_recorder_status.present:
              - name: 'string'
              - resource_id: 'string'
              - recording: 'boolean'

    Returns:
         Dict[str, Any]

    Examples:
        .. code-block:: sls

             aws-config-recorder-status:
               aws.config.config_recorder_status.present:
                 - name: 'config_recorder'
                 - resource_id: 'config_recorder'
                 - recording: true

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if resource_id:
        name = resource_id

    resource = (
        await hub.exec.boto3.client.config.describe_configuration_recorder_status(
            ctx,
            ConfigurationRecorderNames=[name],
        )
    )

    if resource:
        resource_id = name

    if ctx.get("test", False) and "NoSuchConfigurationRecorderException" in str(
        resource["comment"]
    ):
        before = None
    elif not resource["result"]:
        result["comment"] = resource["comment"]
        result["result"] = False
        return result
    else:
        before = resource["ret"]["ConfigurationRecordersStatus"][0]
    if before:
        resource_translated = hub.tool.aws.config.conversion_utils.convert_raw_config_recorder_status_to_present(
            ctx, raw_resource=before
        )

        result["old_state"] = resource_translated
    if ctx.get("test", False):
        result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
            enforced_state={},
            desired_state={
                "name": name,
                "recording": recording,
                "resource_id": resource_id,
            },
        )
        operation = (
            "update" if before is None or recording != before["recording"] else "skip"
        )
        result["comment"] = (
            f"Would {operation} aws.config.config_recorder_status '{name}'",
        )
        return result

    result["new_state"] = copy.deepcopy(result["old_state"])
    if recording != before["recording"]:
        if resource_translated["recording"] == recording:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.config.config_recorder_status", name=name
            )
            result["new_state"] = copy.deepcopy(result["old_state"])
            return result

        if recording:
            recording_status_ret = (
                await hub.exec.boto3.client.config.start_configuration_recorder(
                    ctx=ctx, ConfigurationRecorderName=name
                )
            )
        else:
            recording_status_ret = (
                await hub.exec.boto3.client.config.stop_configuration_recorder(
                    ctx=ctx, ConfigurationRecorderName=name
                )
            )
        if not recording_status_ret["result"]:
            result["comment"] = recording_status_ret["comment"]
            result["result"] = False
            return result
        result["new_state"]["recording"] = recording
        result["comment"] = hub.tool.aws.comment_utils.update_comment(
            resource_type="aws.config.config_recorder_status", name=name
        )
    else:
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.config.config_recorder_status", name=name
        )
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """Reset the configuration recorder status.

    Stops recording configurations of the AWS resources you have selected to record in your AWS account.

    Args:
        name(str):
            The name of the recorder.

        resource_id(str, Optional):
            AWS Config configuration recorder Name. Idem automatically considers this resource being absent
            if this field is not specified.

    Returns:
          Dict[str, Any]

    Examples:
          .. code-block:: sls

            aws-config-recorder-status:
              aws.config.config_recorder_status.absent:
                - name: 'config_recorder'
                - resource_id: 'config_recorder'
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.config.config_recorder_status", name=name
        )
        return result

    resource = (
        await hub.exec.boto3.client.config.describe_configuration_recorder_status(
            ctx,
            ConfigurationRecorderNames=[name],
        )
    )
    if not resource["result"]:
        if "NoSuchConfigurationRecorderException" in str(resource["comment"]):
            result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.config.config_recorder_status", name=name
            )
            return result
        result["comment"] = resource["comment"]
        result["result"] = False
        return result
    before = resource["ret"]["ConfigurationRecordersStatus"][0]

    resource_translated = hub.tool.aws.config.conversion_utils.convert_raw_config_recorder_status_to_present(
        ctx, raw_resource=before
    )
    result["old_state"] = resource_translated
    result["new_state"] = copy.deepcopy(result["old_state"])
    if before:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "recording": False,
                    "resource_id": resource_id,
                },
            )
            result["comment"] = (
                f"Would reset aws.config.config_recorder_status '{name}'",
            )
            return result

        if before["recording"]:
            recording_status_ret = (
                await hub.exec.boto3.client.config.stop_configuration_recorder(
                    ctx=ctx, ConfigurationRecorderName=name
                )
            )
            if not recording_status_ret["result"]:
                result["comment"] = recording_status_ret["comment"]
                result["result"] = False
                return result
            result["new_state"]["recording"] = False
            result["comment"] = (f"Rested aws.config.config_recorder_status '{name}'",)
        else:
            result["comment"] = (
                f"Preserved aws.config.config_recorder_status '{name}'",
            )
    else:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.config.config_recorder_status", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Return details about your config recorder status.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.config.config_recorder_status
    """
    result = {}
    ret = await hub.exec.boto3.client.config.describe_configuration_recorders(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe configuration recorder {ret['comment']}")
        return {}

    for resource in ret["ret"]["ConfigurationRecorders"]:
        name = resource["name"]
        recording_status = (
            await hub.exec.boto3.client.config.describe_configuration_recorder_status(
                ctx,
                ConfigurationRecorderNames=[name],
            )
        )
        if not recording_status["result"]:
            hub.log.warning(
                f"Could not fetch configuration recorder '{name}' status with error {recording_status['comment']}"
            )
            continue

        recorder_name = f"{name}-config-recorder"
        translated_resource = hub.tool.aws.config.conversion_utils.convert_raw_config_recorder_status_to_present(
            ctx,
            raw_resource=resource,
            recording=recording_status["ret"]["ConfigurationRecordersStatus"][0][
                "recording"
            ],
        )
        result[recorder_name] = {
            "aws.config.config_recorder_status.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }
    return result



import os
import pathlib
import shutil

import boto3.session
from dict_tools.data import NamespaceDict

try:
    import tqdm

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


def __init__(hub):
    hub.pop.sub.load_subdirs(hub.pop_create.aws)


def context(hub, ctx, directory: pathlib.Path):
    ctx = hub.pop_create.idem_cloud.init.context(ctx, directory)
    ctx.servers = [None]

    # AWS already has an acct plugin
    ctx.has_acct_plugin = False
    ctx.service_name = "aws_auto"

    if not ctx.get("templates_dir", None):
        # Use default templates
        ctx["templates_dir"] = (
            os.path.dirname(os.path.realpath(__file__))
            + "/{{cookiecutter.root_dir}}/{{cookiecutter.clean_name}}/autogen/{{cookiecutter.service_name}}/templates/"
        )

    # Copy the templates into directory (aka temporary project folder)
    target = f"{directory}/{ctx.clean_name}/autogen/{ctx.service_name}/templates"
    shutil.copytree(ctx.get("templates_dir"), target, dirs_exist_ok=True)

    # Now start getting into AWS resource plugin creation
    session = boto3.session.Session()

    # If CLI provides services then use those services first
    # e.g. --services rds
    services = hub.OPT.pop_create.services or session.get_available_services()
    # This takes a while because we are making http calls to aws
    for aws_service_name in tqdm.tqdm(services, desc="services"):
        # Clean out the service name
        aws_service_name = (
            aws_service_name.lower().strip().replace(" ", "_").replace("-", "_")
        )
        aws_service_name = hub.tool.format.keyword.unclash(aws_service_name)

        # Get supported operations for this service
        resource_operations = hub.pop_create.aws.service.parse_resource_and_operations(
            service_name=aws_service_name,
            session=session,
        )

        requested_service_resources = hub.OPT.pop_create.service_resources
        if bool(requested_service_resources):
            # if the CLI provides resources, filter the list of resources to process
            # e.g. --service_resources db_cluster db_instance
            resource_operations = {
                r: resource_operations[r]
                for r in requested_service_resources
                if r in resource_operations
            }

        resource_modules = {}
        for resource_name, functions in tqdm.tqdm(
            resource_operations.items(), desc="operations"
        ):
            # Clean out resource name
            resource_name = (
                resource_name.lower().strip().replace(" ", "_").replace("-", "_")
            )

            # Check if the plugin should be created
            #   - see if it exists
            #   - or --overwrite flag is used
            resource_plugin_exists = hub.pop_create.aws.init.plugin_exists(
                ctx, aws_service_name, resource_name
            )
            should_create_resource_plugin = (
                ctx.overwrite_existing or not resource_plugin_exists
            )

            if should_create_resource_plugin:
                # parse known or commonly used resource actions for the resource
                resource_functions = hub.pop_create.aws.resource.parse_functions(
                    session,
                    aws_service_name,
                    resource_name,
                    functions,
                )

                # create shared resource data to be used when creating resource plugins
                shared_resource_data = {
                    "aws_service_name": aws_service_name,
                    "resource_name": resource_name,
                    "functions": resource_functions,
                }

                # get resource exec, state, and tool modules
                resource_modules[
                    f"{aws_service_name}.{resource_name}"
                ] = hub.pop_create.aws.plugin.create_resource_plugin(
                    ctx, shared_resource_data
                )

        # There could be lot many resources in a given service, so it generates code for them
        hub.pop_create.aws.init.generate_resource_modules(
            ctx, directory, aws_service_name, resource_modules
        )

        # Now generate any service related modules e.g. tags
        hub.pop_create.aws.init.generate_service_tag_modules(
            ctx, directory, aws_service_name, session
        )

    return ctx


def plugin_exists(hub, ctx, aws_service_name: str, resource_name: str) -> bool:
    """
    Validate if the plugin path exists based on create plugin
    """
    path = pathlib.Path(ctx.target_directory).absolute() / ctx.clean_name
    if "auto_state" in ctx.create_plugin:
        path = path / "exec"
    elif "state_modules" in ctx.create_plugin:
        path = path / "states"
    elif "tests" in ctx.create_plugin:
        path = path / "tests" / "integration" / "states"

    path = path / aws_service_name / f"{resource_name}.py"
    if path.exists():
        hub.log.info(f"Plugin already exists at '{path}', use `--overwrite` to modify")
        return True

    return False


def generate_resource_modules(
    hub, ctx, directory: str, resource_name: str, resource_plugins: dict
):
    try:
        ctx.cloud_spec = NamespaceDict(
            api_version="",
            project_name=ctx.project_name,
            service_name=ctx.service_name,
            request_format={},
            plugins=resource_plugins,
        )

        # pop-create CLI config '--create-plugin' supports 'auto_states', 'exec_modules', 'sls' or 'templates'
        # The following translates it to CloudSpec's create_plugin and runs the code generation from CloudSpec:
        if ctx.create_plugin in ["auto_states", "exec_modules"]:
            hub.cloudspec.init.run(
                ctx,
                directory,
                # exec_modules creates functions for everything except present/absent/describe
                # auto_state creates functions for get/create/update/list/delete
                # tool creates functions for everything which is NOT present/absent/describe/get/create/update/list/delete
                create_plugins=["auto_state", "tool", "tests"],
            )
        elif ctx.create_plugin == "state_modules":
            hub.cloudspec.init.run(
                ctx,
                directory,
                create_plugins=["auto_state", "tool", "state_modules", "sls", "tests"],
            )
        elif ctx.create_plugin == "sls":
            hub.cloudspec.init.run(
                ctx,
                directory,
                create_plugins=["sls"],
            )
        elif ctx.create_plugin == "templates":
            hub.cloudspec.init.run(
                ctx,
                directory,
                create_plugins=["templates"],
            )
        elif ctx.create_plugin == "test_modules":
            hub.cloudspec.init.run(
                ctx,
                directory,
                create_plugins=["tests"],
            )
        else:
            raise ValueError(
                f"Invalid input '{ctx.create_plugin}' for --create-plugin."
            )
    finally:
        hub.log.info(
            f"Finished generating modules for service [{resource_name}] with create plugin {ctx.create_plugin}]"
        )
        # Reset it for the next resource or plugin
        ctx.cloud_spec = None


def generate_service_tag_modules(
    hub,
    ctx,
    directory,
    aws_service_name: str,
    session: "boto3.session.Session",
):
    try:
        # Get service modules e.g. tag
        service_tag_methods = hub.pop_create.aws.service.parse_service_tag_methods(
            session=session,
            aws_service_name=aws_service_name,
        )

        service_tag_plugin = {
            f"{aws_service_name}.tag": hub.pop_create.aws.plugin.create_tags_plugin(
                aws_service_name, service_tag_methods
            )
        }

        ctx.cloud_spec = NamespaceDict(
            api_version="",
            project_name=ctx.project_name,
            service_name=ctx.service_name,
            request_format={},
            plugins=service_tag_plugin,
        )

        # Tags would go under tool modules
        hub.cloudspec.init.run(
            ctx,
            directory,
            create_plugins=["tool"],
        )
    finally:
        hub.log.info(
            f"Finished generating tag modules for service [{aws_service_name}]"
        )
        # Reset it for the next resource or plugin
        ctx.cloud_spec = None



"""Exec module for managing SNS Subscription."""
from typing import Any
from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name: str,
    topic_arn: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Get Subscription from either topic arn or resource id.

    Args:
        name(str):
            An Idem name of the state for logging.

        topic_arn(str, Optional):
            Topic ARN

        resource_id(str, Optional):
            Subscription ARN
    """
    result = dict(comment=[], result=True, ret=None)

    ret = await hub.exec.aws.sns.subscription.list(
        ctx, name=name, topic_arn=topic_arn, resource_id=resource_id
    )
    if not ret["result"]:
        result["result"] = False
        result["comment"] += list(ret["comment"])
        return result

    filtered = ret["ret"]
    result["comment"] = ret["comment"]

    if filtered:
        result["ret"] = filtered[0]
        if len(filtered) > 1:
            result["comment"].append(
                f"More than one aws.sns.subscription resource was found. Use resource {filtered[0].get('resource_id')}"
            )

    return result


async def list_(
    hub, ctx, name: str = None, topic_arn: str = None, resource_id: str = None
) -> Dict:
    """Use an un-managed Subscription as a data-source.

    Args:
        name(str):
            The name of the Idem state.

        topic_arn(str, Optional):
            Topic ARN

        resource_id(str, Optional):
            Subscription ARN

    """
    result = dict(comment=[], result=True, ret=[])
    ret = None
    if topic_arn:
        ret = await hub.exec.boto3.client.sns.list_subscriptions_by_topic(
            ctx, TopicArn=topic_arn
        )
        if not ret["result"]:
            if "NotFoundException" in str(ret["comment"]):
                result["comment"].append(
                    hub.tool.aws.comment_utils.get_empty_comment(
                        resource_type="aws.sns.subscription", name=name
                    )
                )
                result["comment"] += list(ret["comment"])
                return result
            result["result"] = False
            result["comment"] += list(ret["comment"])
            return result
    else:
        ret = await hub.exec.boto3.client.sns.list_subscriptions(ctx)
        if not ret["result"]:
            result["result"] = False
            result["comment"] += list(ret["comment"])
            return result
        if resource_id:
            filtered = list(
                filter(
                    lambda subs: subs["SubscriptionArn"] == resource_id,
                    ret["ret"]["Subscriptions"],
                )
            )
            ret["ret"]["Subscriptions"] = filtered

    if not ret["ret"]["Subscriptions"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.sns.subscription", name=name
            )
        )
        return result

    for subscription in ret["ret"]["Subscriptions"]:
        subscription_id = subscription.get("SubscriptionArn")

        if subscription_id == "Deleted" or subscription_id == "PendingConfirmation":
            continue

        resource = await hub.tool.boto3.resource.create(
            ctx, "sns", "Subscription", arn=subscription_id
        )
        raw_resource = await hub.tool.boto3.resource.describe(resource)

        result["ret"].append(
            hub.tool.aws.sns.conversion_utils.convert_raw_subscription_to_present(
                raw_resource=raw_resource, idem_resource_name=subscription_id
            )
        )

    return result



"""State module for managing backup vault."""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    encryption_key_arn: str = None,
    creator_request_id: str = None,
    tags: Dict[str, str] = None,
) -> Dict[str, Any]:
    """Create/Update the backup vault.

    Creates a logical container where backups are stored. A CreateBackupVault request includes a name, optionally one or
    more resource tags, an encryption key, and a request ID.

    Args:
        name(str):
            An Idem name of the resource.
            The name of a logical container where backups are stored. Backup vaults are identified by names that are
            unique to the account used to create them and the Amazon Web Services Region where they are created.
            They consist of letters, numbers, and hyphens.

        resource_id(str, Optional):
            An Idem resource id of the resource as well as name of the backup vault.

        encryption_key_arn(str, Optional):
            The server-side encryption key that is used to protect your backups;
            for example, arn:aws:kms:us-west-2:111122223333:key/1234abcd-12ab-34cd-56ef-1234567890ab .

        creator_request_id(str, Optional):
            A unique string that identifies the request and allows failed requests to be retried without the risk of
            running the operation twice. This parameter is optional.
            If used, this parameter must contain 1 to 50 alphanumeric or '-_.' characters.

        tags(dict, Optional):
            Metadata that you can assign to help organize the resources that you create. Each tag is a key-value pair.

    Request Syntax:
        .. code-block:: sls

            vault_name:
              aws.backup.backup_vault.present:
              - name: "string"
              - resource_id: "string"
              - encryption_key_arn: "string"
              - creator_request_id: "string"
              - tags: "dict"


    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              aws.backup.backup_vault.present:
                - name: idem_backup_vault
                - resource_id: idem_backup_vault
                - encryption_key_arn: arn:aws:kms:us-east-1:011922870716:key/ff4fd718-612c-4d0f-8172-8150025bb4fa
                - tags:
                    name: backup_vault
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_updated = False

    if resource_id:
        before = await hub.exec.aws.backup.backup_vault.get(
            ctx, resource_id=resource_id, name=name
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.backup.backup_vault", name=name
        )
        result["old_state"] = copy.deepcopy(before["ret"])
        result["new_state"] = copy.deepcopy(result["old_state"])

        if tags is not None and tags != result["old_state"].get("tags"):
            update_ret = await hub.tool.aws.backup.tag.update_tags(
                ctx=ctx,
                backup_vault_arn=result["old_state"].get("backup_vault_arn"),
                old_tags=result["old_state"].get("tags"),
                new_tags=tags,
            )
            result["comment"] = update_ret["comment"]
            if not update_ret["result"]:
                result["result"] = False
                return result

            resource_updated = resource_updated or bool(update_ret["ret"])
            if update_ret["ret"] and ctx.get("test", False):
                result["new_state"].update(update_ret["ret"])
                result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.backup.backup_vault", name=name
                )
            if resource_updated and ctx.get("test", False):
                return result
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "encryption_key_arn": encryption_key_arn,
                    "creator_request_id": creator_request_id,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.backup.backup_vault", name=name
            )
            return result
        create_ret = await hub.exec.boto3.client.backup.create_backup_vault(
            ctx,
            BackupVaultName=name,
            EncryptionKeyArn=encryption_key_arn,
            CreatorRequestId=creator_request_id,
            BackupVaultTags=tags if tags else None,
        )

        if not create_ret["result"]:
            result["result"] = False
            result["comment"] = create_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.backup.backup_vault", name=name
        )
        resource_id = name

    if (not result["old_state"]) or resource_updated:
        after_ret = await hub.exec.aws.backup.backup_vault.get(
            ctx, name=name, resource_id=resource_id
        )
        if not after_ret["result"] or not after_ret["ret"]:
            result["result"] = False
            result["comment"] += tuple(after_ret["comment"])
            return result
        result["new_state"] = after_ret["ret"]

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Deletes the backup vault.

    Deletes the backup vault identified by its name. A vault can be deleted only if it is empty.

    Args:
        name(str): Idem name of the backup vault.
        resource_id(str): Name of the backup vault to be deleted, Idem automatically considers this resource being
            absent if this field is not specified.

    Request Syntax:
        .. code-block:: sls

            vault_name:
              aws.backup.backup_vault.absent:
                - name: "string"
                - resource_id: "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            vault_name:
              aws.backup.backup_vault.absent:
                - name: idem_backup_vault
                - resource_id: idem_backup_vault
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.backup.backup_vault", name=name
        )
        return result

    before = await hub.exec.aws.backup.backup_vault.get(
        ctx, resource_id=resource_id, name=name
    )

    if not before["result"] and not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.backup.backup_vault", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.backup.backup_vault", name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]
        delete_ret = await hub.exec.boto3.client.backup.delete_backup_vault(
            ctx, BackupVaultName=resource_id
        )
        if not delete_ret["result"]:
            result["result"] = False
            result["comment"] = delete_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.backup.backup_vault", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Gets information about the backup vault.

    Returns a list of recovery point storage containers along with information about them.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.backup.backup_Vault
    """

    result = {}
    ret = await hub.exec.aws.backup.backup_vault.list(ctx, "backup_vault")
    if not ret["result"]:
        hub.log.debug(f"Could not describe backup vault {ret['comment']}")
        return {}

    for backup_vault in ret["ret"]:
        vault_name = backup_vault.get("name")

        result[vault_name] = {
            "aws.backup.backup_vault.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in backup_vault.items()
            ]
        }
    return result



import contextlib
import os
import pathlib
import stat
import tempfile
from typing import Tuple

from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.hazmat.primitives.serialization import NoEncryption
from cryptography.hazmat.primitives.serialization import PrivateFormat
from cryptography.hazmat.primitives.serialization import PublicFormat


@contextlib.contextmanager
def verify(
    hub, *, ssh_public_key: str = None, ssh_private_key: str = None
) -> Tuple[str, str]:
    """
    Validate the ssh keypair, generate a keypair if necessary

    Args:
        hub:
        ssh_public_key(Text, Optional): A public ssh key or path to send to the instance
        ssh_private_key(Text, Optional): A private ssh key or path to send to the instance

    Yields:
        A tuple containing the bytes of the public ssh key and the file path of the private ssh key as a tuple

    Examples:
        # Create a new keypair
        with hub.tool.aws.ec2.instance.key_pair.verify() as key_pair:
            verified_public_key, private_key_file = key_pair

        # Validate an existing keypair
        ssh_public_key = "ssh-rsa XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX username@host"
        ssh_private_key = "~/.ssh/id_rsa"
        with hub.tool.aws.ec2.instance.key_pair.verify(ssh_public_key=ssh_public_key, ssh_private_key=ssh_private_key) as key_pair:
            verified_public_key, private_key_file = key_pair
    """
    if bool(ssh_public_key) != bool(ssh_private_key):
        raise ValueError(
            "Both a private ssh key and a public ssh key must be specified"
        )

    # Create a keypair if none were passed
    if not (ssh_public_key and ssh_private_key):
        hub.log.debug(f"Creating an ssh keypair")
        private_key = rsa.generate_private_key(65537, 2048)
        ssh_private_key = private_key.private_bytes(
            encoding=Encoding.PEM,
            format=PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=NoEncryption(),
        )
        ssh_public_key = private_key.public_key().public_bytes(
            encoding=Encoding.OpenSSH, format=PublicFormat.OpenSSH
        )

    # Verify the public key
    if isinstance(ssh_public_key, bytes):
        ssh_public_key = ssh_public_key.decode()

    # If the ssh_public_key was a path, then read the contents of that path into a variable
    try:
        public_key_path = pathlib.Path(ssh_public_key)
        if public_key_path.exists():
            ssh_public_key = public_key_path.read_text()
    except:
        # It's OK, the public key is a string not a file path
        ...

    try:
        # Verify the private key
        private_key_path = pathlib.Path(
            ssh_private_key.decode()
            if isinstance(ssh_private_key, bytes)
            else ssh_private_key
        )
        file_exists = private_key_path.exists()
    except:
        file_exists = False

    if file_exists:
        st_mode = os.stat(private_key_path).st_mode
        if stat.S_ISREG(st_mode) and ((st_mode & 0o777) == 0o600):
            yield ssh_public_key, private_key_path.absolute()
        else:
            raise PermissionError(
                f"Private key has invalid permissions, should be 0o600"
            )
    else:
        with tempfile.NamedTemporaryFile(
            mode="wb+", prefix="id-rsa-", suffix=".pem", delete=True
        ) as private_key_file:
            if isinstance(ssh_private_key, str):
                ssh_private_key = ssh_private_key.encode()
            # Write private_key to file and chmod 400
            private_key_file.write(ssh_private_key)
            private_key_file.flush()
            os.chmod(private_key_file.name, 0o600)
            yield ssh_public_key, private_key_file.name



"""Exec module for managing EC2 VPC Endpoint Service Configurations."""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["soft_fail"]

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    resource_id: str,
    filters: List[
        make_dataclass(
            "Filter",
            [
                ("Name", str, field(default=None)),
                ("Values", List[str], field(default=None)),
            ],
        )
    ] = None,
    name: str = None,
) -> Dict[str, Any]:
    """

    Describes the VPC endpoint service configurations in your account (your services).

    Args:
        resource_id(str): The ID of the service.

        filters(List[dict[str, Any]], Optional): The filters.

            * service-name - The name of the service.

            * service-id - The ID of the service.

            * service-state - The state of the service (Pending | Available | Deleting | Deleted | Failed).

            * supported-ip-address-types - The IP address type (ipv4 | ipv6).

            * tag:<key> - The key/value combination of a tag assigned to the resource. Use the tag key in the filter name and the tag value as the filter value.

            * tag-key - The key of a tag assigned to the resource. Use this filter to find all resources assigned a tag with a specific key, regardless of the tag value. Defaults to None.

            Filter names and Filter values are case-sensitive If you specify multiple values for a
            filter, the values are joined with an OR, and the request returns all results that match any of
            the specified values.

        name(str, Optional): Idem name of the resource. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:
        Resource State:

        .. code-block:: sls

            unmanaged_resource:
              exec.run:
                - path: aws.ec2.vpc_endpoint_service_configuration.get
                - kwargs:
                    resource_id: value

        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.ec2.vpc_endpoint_service_configuration.get resource_id=service-id
    """

    result = dict(comment=[], ret=None, result=True)

    get = await hub.exec.boto3.client.ec2.describe_vpc_endpoint_service_configurations(
        ctx=ctx,
        **{
            "ServiceIds": [resource_id],
            "Filters": filters,
        },
    )

    # Case: Error
    if not get["result"]:
        if "NotFound" in str(get["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.ec2.vpc_endpoint_service_configuration",
                    name=resource_id,
                )
            )
            result["comment"].append(get["comment"])
            return result

        result["comment"].append(get["comment"])
        result["result"] = False
        return result

    # Case: Empty results
    if not get["ret"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ec2.vpc_endpoint_service_configuration",
                name=resource_id,
            )
        )
        return result

    if len(get["ret"]["ServiceConfigurations"]) > 1:
        result["comment"].append(
            hub.tool.aws.comment_utils.find_more_than_one(
                resource_id=resource_id,
                resource_type="aws.ec2.vpc_endpoint_service_configuration",
            )
        )

    raw_resource = get["ret"]["ServiceConfigurations"][0]

    result[
        "ret"
    ] = await hub.tool.aws.ec2.vpc_endpoint_service_configuration.convert_raw_resource_to_present_async(
        ctx=ctx,
        resource_id=resource_id,
        raw_resource=raw_resource,
        idem_resource_name=name,
    )

    return result


async def list_(
    hub,
    ctx,
    service_ids: List[str] = None,
    filters: List[
        make_dataclass(
            "Filter",
            [
                ("Name", str, field(default=None)),
                ("Values", List[str], field(default=None)),
            ],
        )
    ] = None,
) -> Dict[str, Any]:
    """
    Describes the VPC endpoint service configurations in your account (your services).

    Args:
        service_ids(List[str], Optional): The IDs of the endpoint services. Defaults to None.

        filters(List[dict[str, Any]], Optional): The filters.

            * service-name - The name of the service.

            * service-id - The ID of the service.

            * service-state - The state of the service (Pending | Available | Deleting | Deleted | Failed).

            * supported-ip-address-types - The IP address type (ipv4 | ipv6).

            * tag:<key> - The key/value combination of a tag assigned to the resource. Use the tag key in the filter name and the tag value as the filter value.

            * tag-key - The key of a tag assigned to the resource. Use this filter to find all resources assigned a tag with a specific key, regardless of the tag value. Defaults to None.

            Filter names and Filter values are case-sensitive If you specify multiple values for a
            filter, the values are joined with an OR, and the request returns all results that match any of
            the specified values.

    Returns:
        Dict[str, Any]

    Examples:

        Resource State:

        .. code-block:: sls

            unmanaged_resources:
              exec.run:
                - path: aws.ec2.vpc_endpoint_service_configuration.list
                - kwargs:


        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.ec2.vpc_endpoint_service_configuration.list

    """

    result = dict(comment=[], ret=[], result=True)

    ret = await hub.exec.boto3.client.ec2.describe_vpc_endpoint_service_configurations(
        ctx=ctx, **{"ServiceIds": service_ids, "Filters": filters}
    )

    if not ret["result"]:
        result["comment"].append(ret["comment"])
        result["result"] = False
        return result

    if not ret["ret"].get("ServiceConfigurations"):
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.ec2.vpc_endpoint_service_configuration", name=None
            )
        )
        return result

    for resource in ret["ret"]["ServiceConfigurations"]:
        result["ret"].append(
            await hub.tool.aws.ec2.vpc_endpoint_service_configuration.convert_raw_resource_to_present_async(
                ctx=ctx,
                resource_id=resource.get("ServiceId"),
                raw_resource=resource,
                idem_resource_name=None,
            )
        )
    return result


async def create(
    hub,
    ctx,
    name: str = None,
    acceptance_required: bool = None,
    private_dns_name: str = None,
    network_load_balancer_arns: List[str] = None,
    gateway_load_balancer_arns: List[str] = None,
    supported_ip_address_types: List[str] = None,
    client_token: str = None,
    tags: Dict[str, Any] or List = None,
) -> Dict[str, Any]:
    """
    Creates a VPC endpoint service to which service consumers (Amazon Web Services accounts, users, and IAM roles)
    can connect.

    Before you create an endpoint service, you must create one of the following for your service:

        * A Network Load Balancer. Service consumers connect to your service using an interface endpoint.

        * A Gateway Load Balancer. Service consumers connect to your service using a Gateway Load Balancer endpoint.

    If you set the private DNS name, you must prove that you own the private DNS domain name.
    For more information, see the Amazon Web Services PrivateLink Guide.

    Args:
        name(str, Optional): Idem name of the resource. Defaults to None.

        acceptance_required(bool, Optional): Indicates whether requests from service consumers to create an endpoint to your service must be
            accepted manually. Defaults to None.

        private_dns_name(str, Optional): (Interface endpoint configuration) The private DNS name to assign to the VPC endpoint service. Defaults to None.

        network_load_balancer_arns(List[str], Optional): The Amazon Resource Names (ARNs) of the Network Load Balancers. Defaults to None.

        gateway_load_balancer_arns(List[str], Optional): The Amazon Resource Names (ARNs) of the Gateway Load Balancers. Defaults to None.

        supported_ip_address_types(List[str], Optional): The supported IP address types. The possible values are ipv4 and ipv6. Defaults to None.

        client_token(str, Optional): Unique, case-sensitive identifier that you provide to ensure the idempotency of the request. For
            more information, see How to ensure idempotency. Defaults to None.

        tags (Dict or List, Optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of [{"Key": tag-key, "Value": tag-value}] to associate with the VPC. Each tag consists of a key name and an associated value. Defaults to None.

            * Key (str, Optional): The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode characters. May not begin with aws:.
            * Value(str, Optional): The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256 Unicode characters.

    Returns:
        Dict[str, Any]

    Examples:
        Using in a state:

        .. code-block:: sls

            unmanaged_resources:
              exec.run:
                - path: aws.ec2.vpc_endpoint_service_configuration.create
                - kwargs:

        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.ec2.vpc_endpoint_service_configuration.create
    """

    result = dict(comment=[], ret={}, result=True)

    tags = (
        hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
        if isinstance(tags, List)
        else tags
    )

    if not network_load_balancer_arns and not gateway_load_balancer_arns:
        result["result"] = False
        result["comment"].append(
            f"Cannot create aws.ec2.vpc_endpoint_service_configuration as one of network_load_balancer_arn "
            f"or gateway_load_balancer_arn is required",
        )
        return result

    desired_state = {
        k: v
        for k, v in locals().items()
        if k not in ("hub", "ctx", "kwargs", "result") and v is not None
    }

    resource_to_raw_input_mapping = {
        "acceptance_required": "AcceptanceRequired",
        "private_dns_name": "PrivateDnsName",
        "network_load_balancer_arns": "NetworkLoadBalancerArns",
        "gateway_load_balancer_arns": "GatewayLoadBalancerArns",
        "supported_ip_address_types": "SupportedIpAddressTypes",
    }

    payload = {}
    for key, value in desired_state.items():
        if key in resource_to_raw_input_mapping.keys() and value is not None:
            payload[resource_to_raw_input_mapping[key]] = desired_state.get(key)

    if not tags:
        tags = {}

    if "Name" not in tags:
        # Override with input name, if not provided in tags
        tags["Name"] = name

    payload["TagSpecifications"] = [
        {
            "ResourceType": "vpc-endpoint-service",
            "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
        }
    ]

    ret = await hub.exec.boto3.client.ec2.create_vpc_endpoint_service_configuration(
        ctx, ClientToken=client_token or name, **payload
    )

    result["result"] = ret["result"]
    if not result["result"]:
        result["comment"].append(ret["comment"])
        return result

    result["comment"].append(
        f"Created aws.ec2.vpc_endpoint_service_configuration '{name}'",
    )

    raw_resource = ret["ret"].get("ServiceConfiguration", {})
    result["ret"]["resource_id"] = raw_resource.get("ServiceId")
    result["ret"]["name"] = name

    return result


async def update(
    hub,
    ctx,
    resource_id: str,
    name: str = None,
    private_dns_name: str = None,
    remove_private_dns_name: bool = None,
    acceptance_required: bool = None,
    add_network_load_balancer_arns: List[str] = None,
    remove_network_load_balancer_arns: List[str] = None,
    add_gateway_load_balancer_arns: List[str] = None,
    remove_gateway_load_balancer_arns: List[str] = None,
    add_supported_ip_address_types: List[str] = None,
    remove_supported_ip_address_types: List[str] = None,
    tags: Dict[str, Any] or List = None,
) -> Dict[str, Any]:
    """
    Modifies the attributes of your VPC endpoint service configuration. You can change the Network Load Balancers or
    Gateway Load Balancers for your service, and you can specify whether acceptance is required for requests to
    connect to your endpoint service through an interface VPC endpoint.

    If you set or modify the private DNS name, you must prove that you own the private DNS domain name.

    Args:
        resource_id(str): The ID of the service.

        name(str, Optional): Idem name of the resource. Defaults to None.

        private_dns_name(str, Optional): (Interface endpoint configuration) The private DNS name to assign to the endpoint service. Defaults to None.

        remove_private_dns_name(bool, Optional): (Interface endpoint configuration) Removes the private DNS name of the endpoint service. Defaults to None.

        acceptance_required(bool, Optional): Indicates whether requests to create an endpoint to your service must be accepted. Defaults to None.

        add_network_load_balancer_arns(List[str], Optional): The Amazon Resource Names (ARNs) of Network Load Balancers to add to your service configuration. Defaults to None.

        remove_network_load_balancer_arns(List[str], Optional): The Amazon Resource Names (ARNs) of Network Load Balancers to remove from your service
            configuration. Defaults to None.

        add_gateway_load_balancer_arns(List[str], Optional): The Amazon Resource Names (ARNs) of Gateway Load Balancers to add to your service configuration. Defaults to None.

        remove_gateway_load_balancer_arns(List[str], Optional): The Amazon Resource Names (ARNs) of Gateway Load Balancers to remove from your service
            configuration. Defaults to None.

        add_supported_ip_address_types(List[str], Optional): The IP address types to add to your service configuration. Defaults to None.

        remove_supported_ip_address_types(List[str], Optional): The IP address types to remove from your service configuration. Defaults to None.

        tags (Dict or List, Optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of [{"Key": tag-key, "Value": tag-value}] to associate with the VPC. Each tag consists of a key name and an associated value. Defaults to None.

            * Key (str, Optional): The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode characters. May not begin with aws:.
            * Value(str, Optional): The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256 Unicode characters.

    Returns:
        Dict[str, Any]

    Examples:
        Using in a state:

        .. code-block:: sls

            unmanaged_resources:
              exec.run:
                - path: aws.ec2.vpc_endpoint_service_configuration.update
                - kwargs:
                    - resource_id: value

        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.ec2.vpc_endpoint_service_configuration.update resource_id=value
    """

    result = dict(comment=[], ret={}, result=True)

    tags = (
        hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
        if isinstance(tags, List)
        else tags
    )

    desired_state = {
        k: v
        for k, v in locals().items()
        if k not in ("hub", "ctx", "kwargs", "result") and v is not None
    }

    payload = {}

    resource_to_raw_input_mapping = {
        "private_dns_name": "PrivateDnsName",
        "remove_private_dns_name": "RemovePrivateDnsName",
        "acceptance_required": "AcceptanceRequired",
        "add_network_load_balancer_arns": "AddNetworkLoadBalancerArns",
        "remove_network_load_balancer_arns": "RemoveNetworkLoadBalancerArns",
        "add_gateway_load_balancer_arns": "AddGatewayLoadBalancerArns",
        "remove_gateway_load_balancer_arns": "RemoveGatewayLoadBalancerArns",
        "add_supported_ip_address_types": "AddSupportedIpAddressTypes",
        "remove_supported_ip_address_types": "RemoveSupportedIpAddressTypes",
    }

    for key, value in desired_state.items():
        if key in resource_to_raw_input_mapping.keys() and value is not None:
            payload[resource_to_raw_input_mapping[key]] = desired_state.get(key)

    if payload:
        payload["ServiceId"] = resource_id

        ret = await hub.exec.boto3.client.ec2.modify_vpc_endpoint_service_configuration(
            ctx, **payload
        )

        if not ret["result"] or not ret["ret"].get("Return"):
            result["result"] = False
            result["comment"].append(
                f"Could not update aws.ec2.vpc_endpoint_service_configuration '{name}'",
            )
            result["comment"].append(ret["comment"])
            return result

        result["comment"].append(
            f"Updated aws.ec2.vpc_endpoint_service_configuration '{name}'",
        )

        get_tags_ret = await hub.tool.aws.ec2.tag.get_tags_for_resource(
            ctx, resource_id=resource_id
        )

        if not tags:
            tags = {}

        if "Name" in get_tags_ret["ret"] and "Name" not in tags:
            # For this resource names are added in tags
            tags["Name"] = get_tags_ret["ret"]["Name"]

        if get_tags_ret["result"]:
            current_tags = get_tags_ret.get("ret", {})
            update_tags_ret = await hub.tool.aws.ec2.tag.update_tags(
                ctx, resource_id=resource_id, old_tags=current_tags, new_tags=tags
            )
            if not update_tags_ret["result"]:
                result["result"] = False
                result["comment"] += update_tags_ret["comment"]
                return result

        result["ret"]["resource_id"] = resource_id
        result["ret"]["name"] = name

    return result


async def delete(hub, ctx, resource_id: str, name: str = None) -> Dict[str, Any]:
    """
    Deletes the specified VPC endpoint service configurations. Before you can delete an endpoint service
    configuration, you must reject any Available or PendingAcceptance interface endpoint connections that are
    attached to the service.

    Args:
        resource_id(str): The ID of the services.

        name(str, Optional): Idem name of the resource. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:
        Resource State:

        .. code-block:: sls

            resource_is_absent:
              aws.ec2.vpc_endpoint_service_configuration.absent:
                - resource_id: value

        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.ec2.vpc_endpoint_service_configuration.delete resource_id=value
    """

    result = dict(comment=[], ret=[], result=True)

    delete_ret = (
        await hub.exec.boto3.client.ec2.delete_vpc_endpoint_service_configurations(
            ctx, **{"ServiceIds": [resource_id]}
        )
    )

    result["result"] = delete_ret["result"]

    if not result["result"]:
        result["comment"] = delete_ret["comment"]
        result["result"] = False
        return result

    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
        resource_type="aws.ec2.vpc_endpoint_service_configuration", name=name
    )

    return result



"""Util functions for SNS topic policy."""
import copy
import json
from typing import Any
from typing import Dict
from typing import List

from dict_tools import differ


def get_default_topic_policy(hub, topic_arn: str) -> str:
    """Topic policy initialized with default attribute values.

    Args:
        topic_arn(str): AWS ARN of the SNS topic

    Returns:
        Dict(str, Any)
    """
    default_policy = {
        "Version": "2008-10-17",
        "Id": "__default_policy_ID",
        "Statement": [
            {
                "Sid": "__default_statement_ID",
                "Effect": "Allow",
                "Principal": {"AWS": "*"},
                "Action": [
                    "SNS:GetTopicAttributes",
                    "SNS:SetTopicAttributes",
                    "SNS:AddPermission",
                    "SNS:RemovePermission",
                    "SNS:DeleteTopic",
                    "SNS:Subscribe",
                    "SNS:ListSubscriptionsByTopic",
                    "SNS:Publish",
                ],
                "Resource": "topic_arn",
                "Condition": {"StringEquals": {"AWS:SourceOwner": "acct_id"}},
            }
        ],
    }
    default_policy["Statement"][0]["Resource"] = topic_arn
    acct_id = topic_arn.split(":")[4]
    default_policy["Statement"][0]["Condition"]["StringEquals"][
        "AWS:SourceOwner"
    ] = acct_id
    policy = json.dumps(default_policy, separators=(", ", ": "))
    return policy


async def update_topic_attributes(
    hub,
    ctx,
    resource_arn: str,
    old_attributes: Dict[str, str],
    new_attributes: Dict[str, str],
):
    """Given old and new attributes of SNS topic the function updates the attributes.

    Args:
        resource_arn(str):
            AWS resource_arn of SNS topic

        old_attributes(Dict):
            old attribute of SNS topic

        new_attributes(Dict):
            new attribute of SNS topic

    Returns:
        .. code-block:: json

            {"result": True|False, "comment": A message Tuple, "ret": Dict}
    """
    result = dict(comment=(), result=True, ret=None)
    attributes_diff = differ.deep_diff(old_attributes, new_attributes)
    attributes_to_update = attributes_diff.get("new")

    if attributes_to_update:
        if ctx.get("test", False):
            result["ret"] = {"updated_attributes": attributes_to_update}
            result["comment"] = (
                f"Would Update attributes {attributes_to_update.keys()}",
            )
            return result

        else:
            for key, value in attributes_to_update.items():
                ret = await hub.exec.boto3.client.sns.set_topic_attributes(
                    ctx, TopicArn=resource_arn, AttributeName=key, AttributeValue=value
                )
                if not ret["result"]:
                    result["comment"] = ret["comment"]
                    result["result"] = False
                    return result

            result["ret"] = {"updated_attributes": attributes_to_update}
            result["comment"] = (f"Updated attributes {attributes_to_update.keys()}",)
    return result


async def update_subscription_attributes(
    hub,
    ctx,
    resource_arn: str,
    old_attributes: Dict[str, str],
    new_attributes: Dict[str, str],
):
    """Given old and new attributes of SNS topic_subscription the function updates the attributes.

    Args:
        resource_arn(str):
            AWS resource_arn of SNS topic_subscription

        old_attributes(Dict):
            old attribute of SNS topic_subscription

        new_attributes(Dict):
            new attribute of SNS topic_subscription

    Returns:
        .. code-block:: json

            {"result": True|False, "comment": A message Tuple, "ret": Dict}
    """
    result = dict(comment=(), result=True, ret=None)
    attributes_diff = differ.deep_diff(old_attributes, new_attributes)
    attributes_to_update = attributes_diff.get("new")

    if attributes_to_update:
        if ctx.get("test", False):
            result["ret"] = {"updated_attributes": attributes_to_update}
            result["comment"] = (
                f"Would Update attributes {attributes_to_update.keys()}",
            )
            return result

        else:
            for key, value in attributes_to_update.items():
                ret = await hub.exec.boto3.client.sns.set_subscription_attributes(
                    ctx,
                    SubscriptionArn=resource_arn,
                    AttributeName=key,
                    AttributeValue=value,
                )
                if not ret["result"]:
                    result["comment"] = ret["comment"]
                    result["result"] = False
                    return result

            result["ret"] = {"updated_attributes": attributes_to_update}
            result["comment"] = (f"Updated attributes {attributes_to_update.keys()}",)
    return result


async def update_tags(
    hub,
    ctx,
    resource_arn,
    old_tags: List[Dict[str, Any]] or Dict[str, Any] = None,
    new_tags: List[Dict[str, Any]] or Dict[str, Any] = None,
):
    """Update tags of AWS SNS resources.

    Args:
        resource_arn(str):
            aws resource arn

        old_tags(List, Optional):
            list of old tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
            {tag-key: tag-value}

        new_tags(List, Optional):
            list of new tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
            {tag-key: tag-value}

    Returns:
        .. code-block:: json

            {"result": True|False, "comment": A message Tuple, "ret": Dict}
    """
    tags_to_add = {}
    tags_to_remove = {}
    if isinstance(old_tags, List):
        old_tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(old_tags)
    if isinstance(new_tags, List):
        new_tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(new_tags)
    if new_tags is not None:
        tags_to_remove, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )
    result = dict(comment=(), result=True, ret={})
    if (not tags_to_remove) and (not tags_to_add):
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result
    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.sns.untag_resource(
                ctx, ResourceArn=resource_arn, TagKeys=list(tags_to_remove)
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result
    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.sns.tag_resource(
                ctx,
                ResourceArn=resource_arn,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags_to_add),
            )
            if not add_ret["result"]:
                result["comment"] = add_ret["comment"]
                result["result"] = False
                return result
    result["ret"] = new_tags
    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    else:
        result["comment"] = hub.tool.aws.comment_utils.update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    return result



"""Exec module for managing AWS WAFV2 IPSet resources."""
__func_alias__ = {"list_": "list"}


async def get(hub, ctx, name: str, scope: str, resource_id: str):
    """Get an IPSets resource from AWS WAFV2.

    Get a single IPSet from AWS. The function returns None when no resource is found.

    Args:
        name(str):
            This is the name of the IPSet in AWS.

        scope(str):
            Specifies whether this is for an Amazon CloudFront distribution or for a regional application.
            Scope is required to find the resource id for the given IPSet.
            For regional applications, you can use any of the endpoints in the list. A regional application can be an
            Application Load Balancer (ALB), an Amazon API Gateway REST API, an AppSync GraphQL API, or an Amazon
            Cognito user pool. For Amazon CloudFront applications, you must use the API endpoint listed for US East (N. Virginia): us-east-1.

        resource_id(str):
            It is a unique identifier for the ip set in AWS.

    Examples:
        Calling from the CLI:

        .. code-block:: bash

            idem exec aws.wafv2.ip_set.get name= "unmanaged_ipsets" scope="REGIONAL" resource_id="resource_id"

        Calling this exec module function from within a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.wafv2.ip_set.get
                - kwargs:
                    name: unmanaged-resource-name
                    scope: scope-name
                    resource_id: aws-resource-id
    """
    result = dict(comment=[], ret=None, result=True)
    get_response = await hub.exec.boto3.client.wafv2.get_ip_set(
        ctx, Name=name, Scope=scope, Id=resource_id
    )
    if not get_response["result"]:
        if "WAFNonexistentItemException" in str(get_response["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.wafv2.ip_set", name=name
                )
            )
            result["comment"] += list(get_response["comment"])
            return result
        result["result"] = False
        result["comment"] = list(get_response["comment"])
        return result
    response = get_response["ret"]
    arn = response["IPSet"]["ARN"]
    tags_ret = await hub.tool.aws.wafv2.tag.get_tags_for_resource(ctx, resource_arn=arn)
    tags = tags_ret["ret"]
    if not tags_ret["result"]:
        result["result"] = False
        result["comment"] += list(tags_ret["comment"])
        return result
    result["ret"] = hub.tool.aws.wafv2.conversion_utils.convert_raw_ip_set_to_present(
        scope=scope,
        raw_resource=get_response,
        tags=tags,
    )
    return result


async def list_(hub, ctx, scope: str):
    """Get the list of IPSets resources from AWS WAFV2.

    Args:
        scope(str):
            Specifies whether this is for an Amazon CloudFront distribution or for a regional application.
            Scope is required to find the resource id for the given IPSet. For regional applications, you can use any of
            the endpoints in the list. A regional application can be an Application Load Balancer (ALB), an Amazon API
            Gateway REST API, an AppSync GraphQL API, or an Amazon Cognito user pool. For Amazon CloudFront
            applications, you must use the API endpoint listed for US East (N. Virginia): us-east-1.

    Examples:
        Calling from the CLI:

        .. code-block:: bash

            idem exec aws.wafv2.ip_set.list scope="REGIONAL"

        Calling this exec module function from within a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.wafv2.ip_set.list
                - kwargs:
                    scope: scope-name
    """
    result = dict(comment=[], ret=[], result=True)
    ret = await hub.exec.boto3.client.wafv2.list_ip_sets(ctx, Scope=scope)
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not "IPSets" in ret["ret"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.wafv2.ip_set", name=scope
            )
        )
        return result
    for ipset in ret["ret"]["IPSets"]:
        idem_resource_name = ipset["Name"]
        resource_id = ipset["Id"]
        raw_resource = await hub.exec.aws.wafv2.ip_set.get(
            ctx, name=idem_resource_name, scope=scope, resource_id=resource_id
        )
        if not raw_resource["result"]:
            result["comment"] += list(ret["comment"])
            result["result"] = False
            return result
        result["ret"].append(raw_resource)
    return result



"""Tool module for managing Amazon S3 bucket website."""
from typing import Any
from typing import Dict


def convert_raw_bucket_website_to_present(
    hub, bucket: str, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    """Convert an AWS S3 bucket website resource to a common idem present state.

    Args:
        bucket(str):
            The S3 bucket name in Amazon Web Services.

        raw_resource(Dict[str, Any]):
            The AWS response to convert.

        idem_resource_name(str, Optional):
            An Idem name of the resource.

    Returns:
        Dict[str, Any]: Common idem present state.
    """
    raw_resource.pop("ResponseMetadata", None)

    resource_translated = {
        "name": idem_resource_name if idem_resource_name else f"{bucket}-website",
        "resource_id": bucket,
        "bucket": bucket,
        "website_configuration": raw_resource,
    }

    return resource_translated



from typing import Dict
from typing import List

__func_alias__ = {"list_": "list"}
__contracts__ = ["soft_fail"]


async def get(
    hub,
    ctx,
    name: str = None,
    resource_id: str = None,
    filters: List[Dict[str, str]] = None,
) -> Dict:
    """
    Get a Network Interface resource from AWS. Supply one of the inputs as the filter.

    Args:
        name(str):
            The name of the Idem state.

        resource_id(str, Optional):
            AWS Network Interface id to identify the resource.

        filters(list, Optional):
            One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
             https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_network_interfaces
    """
    result = dict(comment=[], ret=None, result=True)
    if filters:
        filters = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
            filters=filters
        )

    if resource_id:
        ret = await hub.exec.boto3.client.ec2.describe_network_interfaces(
            ctx,
            NetworkInterfaceIds=[resource_id],
            Filters=filters,
        )
    else:
        ret = await hub.exec.boto3.client.ec2.describe_network_interfaces(
            ctx,
            Filters=filters,
        )

    if not ret["result"]:
        if "InvalidNetworkInterfaceID.NotFound" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.ec2.network_interface", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    present_states = hub.tool.aws.ec2.network_interface.convert_to_present(ret.ret)

    # If the resource can't be found but there were no results then "result" is True and "ret" is None
    if not present_states:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.ec2.network_interface", name=name
            )
        )
        return result

    # return the first result as a plain dictionary
    result["ret"] = next(iter((present_states).values()))

    return result


async def list_(hub, ctx, name: str = None, filters: List = None) -> Dict:
    """
    Use an un-managed Network Interface as a data-source. Supply one of the inputs as the filter.

    Args:
        name(str, Optional):
            The name of the Idem state.

        filters(list, Optional):
            One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_network_interfaces
    """

    result = dict(comment=[], ret=[], result=True)
    if filters:
        filters = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
            filters=filters
        )
    ret = await hub.exec.boto3.client.ec2.describe_network_interfaces(
        ctx,
        Filters=filters,
    )

    # If there was an error in the call then report failure
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    present_states = hub.tool.aws.ec2.network_interface.convert_to_present(ret.ret)
    if not present_states:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.ec2.network_interface", name=name
            )
        )
        return result

    # Return a list of dictionaries with details about all the instances
    result["ret"] = list(present_states.values())

    return result


async def create(
    hub,
    ctx,
    *,
    name: str = None,
    client_token: str = None,
    subnet_id: str,
    description: str = None,
    groups: List[str] = None,
    interface_type: str = None,
    primary_ip_address: str = None,
    private_ip_addresses: List[str] = None,
    ipv4_address_count: int = None,
    ipv4_prefix_count: int = None,
    ipv4_prefixes: List[str] = None,
    ipv6_addresses: List[str] = None,
    ipv6_address_count: int = None,
    ipv6_prefix_count: int = None,
    ipv6_prefixes: List[str] = None,
    tags: Dict[str, str] = None,
    **kwargs,
):
    """
    Create a Network Interface
    """
    result = dict(comment=[], ret=[], result=True)

    # "describe" will define multiple of these parameters, some of which are mutually exclusive.
    validated_ipv6_addresses = None
    validated_ipv6_prefixes = None
    validated_ipv6_address_count = None
    validated_ipv4_addresses = None
    validated_ipv4_prefixes = None
    validated_ipv4_address_count = None

    # Default to specific addresses, fallback to prefixes, then count
    if ipv6_addresses:
        validated_ipv6_addresses = [{"IPv6Address": i} for i in ipv6_addresses]
    elif ipv6_prefixes:
        validated_ipv6_prefixes = [{"IPv6Prefix": i} for i in ipv6_prefixes]
    elif ipv6_address_count:
        validated_ipv6_address_count = ipv6_address_count

    # Add primary IP in the list of IPs, if it's not already there
    if primary_ip_address:
        if not private_ip_addresses:
            private_ip_addresses = [primary_ip_address]
        elif primary_ip_address not in private_ip_addresses:
            private_ip_addresses.append(primary_ip_address)

    # Default to specific addresses, fallback to prefixes, then count
    if private_ip_addresses:
        if not primary_ip_address:
            result["comment"].append(
                "When private_ip_addresses is used, the primary address should also be specified in primary_ip_address"
            )
            result["result"] = False
            return result

        validated_ipv4_addresses = [
            {"PrivateIpAddress": i, "Primary": primary_ip_address == i}
            for i in private_ip_addresses
        ]
    elif ipv4_prefixes:
        validated_ipv4_prefixes = [{"Ipv4Prefix": i} for i in ipv4_prefixes]
    elif ipv4_address_count:
        validated_ipv4_address_count = ipv4_address_count

    tag_specification = None
    if tags:
        tag_specification = [
            {
                "ResourceType": "network-interface",
                "Tags": [{"Key": k, "Value": v} for k, v in tags.items()],
            }
        ]

    ret = await hub.exec.boto3.client.ec2.create_network_interface(
        ctx,
        **{
            "Description": description,
            "Groups": groups,
            "Ipv6Addresses": validated_ipv6_addresses,
            # If concrete IPs are specified in the request, all IPs are included in this list, even the primary IP
            "PrivateIpAddresses": validated_ipv4_addresses,
            "Ipv4Prefixes": validated_ipv4_prefixes,
            "Ipv6Prefixes": validated_ipv6_prefixes,
            "SecondaryPrivateIpAddressCount": validated_ipv4_address_count,
            "Ipv6AddressCount": validated_ipv6_address_count,
            "Ipv4PrefixCount": ipv4_prefix_count,
            "Ipv6PrefixCount": ipv6_prefix_count,
            "InterfaceType": interface_type,
            "SubnetId": subnet_id,
            "TagSpecifications": tag_specification,
            "ClientToken": client_token,
        },
    )
    result["result"] = ret["result"]
    if not result["result"]:
        result["comment"] = ret["comment"]
        return result

    result["comment"] += [f"Created '{name}'"]
    resource_id = ret.ret["NetworkInterface"]["NetworkInterfaceId"]
    result["new_state"] = dict(name=name, resource_id=resource_id)

    present_ret = hub.tool.aws.ec2.network_interface.convert_to_present(
        describe_network_interfaces={"NetworkInterfaces": [ret.ret["NetworkInterface"]]}
    )
    result["ret"] = present_ret[resource_id]
    return result


async def update(
    hub,
    ctx,
    **kwargs,
):
    """
    Update a Network Interface
    """
    result = dict(comment=[], ret=[], result=True)
    return result


async def delete(hub, ctx, *, name: str = None, resource_id: str, **kwargs):
    """
    Delete a Network Interface
    """
    result = dict(comment=[], ret=[], result=True)
    ret = await hub.exec.boto3.client.ec2.delete_network_interface(
        ctx, **{"NetworkInterfaceId": resource_id}
    )
    result["result"] = ret["result"]
    if not result["result"]:
        result["comment"] = ret["comment"]
        result["result"] = False
        return result
    result["comment"] += (f"Deleted '{name}'",)

    return result



from collections import OrderedDict
from typing import Any
from typing import Dict


async def update_security_group_rule(
    hub,
    ctx,
    before: Dict[str, Any],
    input_params: Dict[str, Any],
    resource_id: str,
):
    """Updates the Security group rule.

    Args:
        before(dict[str, Any]): existing resource
        input_params(dict[str, Any]): a dictionary with newly passed values of params.
        resource_id(str): AWS Security group rule ID.

    Returns:
        .. code-block:: json

           {
             "result": True|False,
             "comment": A message Tuple,
             "ret": Dict
           }

    """
    result = dict(comment=(), result=True, ret=None)
    updated_payload = {}
    updated_rule = {}
    new_modified_rule = dict(SecurityGroupRuleId=None, SecurityGroupRule=None)
    new_modified_rule["SecurityGroupRuleId"] = resource_id
    resource_parameters = OrderedDict(
        {
            "IpProtocol": "ip_protocol",
            "FromPort": "from_port",
            "ToPort": "to_port",
            "CidrIpv4": "cidr_ipv4",
            "CidrIpv6": "cidr_ipv6",
            "PrefixListId": "prefix_list_id",
            "Description": "description",
        }
    )

    for key, value in resource_parameters.items():
        if value in before.keys():
            updated_payload[key] = input_params.get(value)
            if (input_params.get(value) is not None) and (
                input_params.get(value) != before[value]
            ):
                updated_rule[key] = input_params[value]
        elif input_params.get(value):
            updated_payload[key] = input_params[value]
            updated_rule[key] = input_params[value]
    # You must either specify cidr_ipv4 or referenced_group_info. You cannot change source from cidr to reference
    # group id or vice versa
    if before.get(
        "cidr_ipv4"
    ) is None and not hub.tool.aws.state_comparison_utils.compare_dicts(
        input_params.get("referenced_group_info"),
        before.get("referenced_group_info"),
    ):
        updated_rule["ReferencedGroupId"] = input_params.get(
            "referenced_group_info"
        ).get("GroupId")
    if updated_rule:
        if (
            before.get("cidr_ipv4") is None
            and input_params.get("referenced_group_info") is not None
        ):
            updated_payload["ReferencedGroupId"] = input_params.get(
                "referenced_group_info"
            ).get("GroupId")
        new_modified_rule["SecurityGroupRule"] = updated_payload
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.ec2.modify_security_group_rules(
                ctx,
                GroupId=input_params.get("group_id"),
                SecurityGroupRules=[new_modified_rule],
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
                return result
        result["ret"] = {}
        result = update_result(result, updated_rule)

    return result


def update_result(
    result: Dict[str, Any], update_payload: Dict[str, Any]
) -> Dict[str, Any]:
    parameters = OrderedDict(
        {
            "IpProtocol": "ip_protocol",
            "FromPort": "from_port",
            "ToPort": "to_port",
            "CidrIpv4": "cidr_ipv4",
            "CidrIpv6": "cidr_ipv6",
            "PrefixListId": "prefix_list_id",
            "Description": "description",
            "ReferencedGroupId": "referenced_group_id",
        }
    )

    for raw_parameter, present_parameter in parameters.items():
        if raw_parameter in update_payload:
            result["ret"][present_parameter] = update_payload[raw_parameter]
            result["comment"] = result["comment"] + (
                f"Update {present_parameter}: {update_payload[raw_parameter]}",
            )
    return result



"""Exec module for managing managing ACM certificates."""
from typing import Dict


async def get(
    hub,
    ctx,
    name: str,
    resource_id: str,
) -> Dict:
    """Retrieves the specified ACM certificate.

    Use an un-managed ACM certificate as a data-source.

    Args:
        name(str):
            The name of the Idem state.

        resource_id(str):
            The Amazon Resource Name (ARN) of certificate to identify the resource.

    Returns:
        Dict[str, Any]:
            Returns certificate_manager in present format

    Examples:
        Calling this exec module function from the cli with resource_id

        .. code-block:: bash

            idem exec aws.acm.certificate_manager.get name="unmanaged_certificate" resource_id="resource_id"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.acm.certificate_manager.get
                - kwargs:
                    name: my_resource
                    resource_id: resource_id

    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.exec.boto3.client.acm.describe_certificate(
        ctx, CertificateArn=resource_id
    )

    if not ret["result"]:
        if "ResourceNotFoundException" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.acm.certificate_manager", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    if ret["ret"]:
        result = (
            await hub.tool.aws.acm.conversion_utils.convert_raw_acm_to_present_async(
                ctx,
                raw_resource=ret["ret"].get("Certificate"),
                idem_resource_name=name,
            )
        )

    return result



"""Exec module for managing Guardduty members."""
from typing import Any
from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    detector_id: str,
    account_id: str,
    resource_id: str = None,
    name: str = None,
) -> Dict[str, Any]:
    """Get info about AWS Guardduty member based on resource_id which is the combination of detector_id:account_id.

    Args:
        detector_id(str):
            AWS Guardduty Detector id

        account_id(str):
            AWS Guardduty member account_id

        resource_id(str, Optional):
            Combination of AWS Guardduty Detector id and AWS Guardduty Member account id. Format:detector_id:account_id

        name(str, Optional):
            Name of the Idem state

    Returns:
        Dict[str, Any]:
            Returns guardduty member in present format

    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec aws.guardduty.member.get name="get member" detector_id="detector_id" account_id="account_id" resource_id="detector_id:account_id"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resources:
              exec.run:
                - path: aws.guardduty.member.get
                - kwargs:
                    name: "get Guardduty member"
                    detector_id: "7ec2638373ab39e421bbc97333706434"
                    account_id: "257230699585"
                    resource_id: "7ec2638373ab39e421bbc97333706434:257230699585 "
    """
    result = dict(comment=[], result=True, ret=None)

    if resource_id and ":" in resource_id:
        extracted_detector_id = resource_id.split(":")[0]
        extracted_account_id = resource_id.split(":")[1]
        if extracted_detector_id != detector_id:
            result["result"] = False
            result[
                "comment"
            ] = f"Detector_id {extracted_detector_id} from resource_id and detector_id {detector_id} passed as arg do not match."
            return result
        elif extracted_account_id != account_id:
            result["result"] = False
            result[
                "comment"
            ] = f"Account_id {extracted_account_id} from resource_id and account_id {account_id} passed as arg do not match."
            return result
        else:
            detector_id = extracted_detector_id
            account_id = extracted_account_id
    account_ids = [account_id]
    ret = await hub.exec.boto3.client.guardduty.get_members(
        ctx, DetectorId=detector_id, AccountIds=account_ids
    )

    if not ret["result"]:
        result["result"] = ret["result"]
        result["comment"] += list(ret["comment"])
        return result

    members = ret["ret"]["Members"]
    if not members:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.guardduty.member", name=name
            )
        )
        return result
    resource = members[0]
    if len(members) > 1:
        result["comment"].append(
            f"More than one aws.guardduty.member resource was found. Use resource {resource.get('AccountId')}"
        )
    result[
        "ret"
    ] = hub.tool.aws.guardduty.conversion_utils.convert_raw_member_to_present(
        member=resource,
        idem_resource_name=name,
        detector_id=detector_id,
    )

    return result


async def list_(
    hub, ctx, detector_id: str, name: str = None, only_associated: str = "False"
) -> Dict[str, Any]:
    """List AWS Guardduty members for a detector.

    Args:
        detector_id(str):
            AWS Guardduty Detector id

        name(str, Optional):
            Name of the Idem state.

        only_associated(str, Optional):
            Specifies whether to only return associated members or return all members.
            All members will include members that haven't been invited or have been disassociated."

    Returns:
        Dict[str, Any]:
            Returns guardduty members in present format

    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec aws.guardduty.member.list name="list members" detector_id="detector_id" only_associated="False"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resources:
              exec.run:
                - path: aws.guardduty.member.list
                - kwargs:
                    name: my_resources
                    detector_id: "a6c242807703f20b65afd91d54dc8fb2"
                    only_associated: 'False'
    """
    result = dict(comment=[], result=True, ret=[])
    ret = await hub.exec.boto3.client.guardduty.list_members(
        ctx, DetectorId=detector_id, OnlyAssociated=only_associated
    )
    if not ret["result"]:
        result["result"] = ret["result"]
        result["comment"] += list(ret["comment"])
        return result

    if not ret["ret"]["Members"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.guardduty.member", name=name
            )
        )
        return result
    # list_members returns the same response as get_members so directly passing it to conversion_utils
    for member in ret["ret"]["Members"]:
        result["ret"].append(
            hub.tool.aws.guardduty.conversion_utils.convert_raw_member_to_present(
                member=member,
                idem_resource_name=name,
                detector_id=detector_id,
            )
        )
    return result



"""State module for managing Cloudfront Cache policies."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    default_ttl: int,
    max_ttl: int,
    min_ttl: int,
    parameter_in_cache_key_and_forward_to_origin: make_dataclass(
        "ParametersInCacheKeyAndForwardedToOrigin",
        [
            ("EnableAcceptEncodingGzip", bool),
            (
                "HeadersConfig",
                make_dataclass(
                    "CachePolicyHeadersConfig",
                    [
                        ("HeaderBehavior", str),
                        (
                            "Headers",
                            make_dataclass(
                                "Headers",
                                [
                                    ("Quantity", int),
                                    ("Items", List[str], field(default=None)),
                                ],
                            ),
                            field(default=None),
                        ),
                    ],
                ),
            ),
            (
                "CookiesConfig",
                make_dataclass(
                    "CachePolicyCookiesConfig",
                    [
                        ("CookieBehavior", str),
                        (
                            "Cookies",
                            make_dataclass(
                                "CookieNames",
                                [
                                    ("Quantity", int),
                                    ("Items", List[str], field(default=None)),
                                ],
                            ),
                            field(default=None),
                        ),
                    ],
                ),
            ),
            (
                "QueryStringsConfig",
                make_dataclass(
                    "CachePolicyQueryStringsConfig",
                    [
                        ("QueryStringBehavior", str),
                        (
                            "QueryStrings",
                            make_dataclass(
                                "QueryStringNames",
                                [
                                    ("Quantity", int),
                                    ("Items", List[str], field(default=None)),
                                ],
                            ),
                            field(default=None),
                        ),
                    ],
                ),
            ),
            ("EnableAcceptEncodingBrotli", bool, field(default=None)),
        ],
    ) = None,
    comment: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Creates a cache policy.

    After you create a cache policy, you can attach it to one or more cache behaviors. When
    it’s attached to a cache behavior, the cache policy determines the following:   The values that CloudFront
    includes in the cache key. These values can include HTTP headers, cookies, and URL query strings. CloudFront
    uses the cache key to find an object in its cache that it can return to the viewer.   The default, minimum, and
    maximum time to live (TTL) values that you want objects to stay in the CloudFront cache.   The headers, cookies,
    and query strings that are included in the cache key are automatically included in requests that CloudFront
    sends to the origin. CloudFront sends a request when it can’t find an object in its cache that matches the
    request’s cache key. If you want to send values to the origin but not include them in the cache key, use
    OriginRequestPolicy. For more information about cache policies, see Controlling the cache key in the Amazon
    CloudFront Developer Guide.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, optional):
            An identifier of the resource in the provider. Defaults to None.

        cache_policy_config(dict[str, Any]):
            A cache policy configuration.

        comment (str, optional):
            A comment to describe the cache policy. The comment cannot be longer than 128 characters.

        default_ttl (int, optional):
            The default amount of time, in seconds, that you want objects to stay in the CloudFront cache
            before CloudFront sends another request to the origin to see if the object has been updated.
            CloudFront uses this value as the object’s time to live (TTL) only when the origin does not send
            Cache-Control or Expires headers with the object. For more information, see Managing How Long
            Content Stays in an Edge Cache (Expiration) in the Amazon CloudFront Developer Guide. The
            default value for this field is 86400 seconds (one day). If the value of MinTTL is more than
            86400 seconds, then the default value for this field is the same as the value of MinTTL.

        max_ttl (int, optional):
            The maximum amount of time, in seconds, that objects stay in the CloudFront cache before
            CloudFront sends another request to the origin to see if the object has been updated. CloudFront
            uses this value only when the origin sends Cache-Control or Expires headers with the object. For
            more information, see Managing How Long Content Stays in an Edge Cache (Expiration) in the
            Amazon CloudFront Developer Guide. The default value for this field is 31536000 seconds (one
            year). If the value of MinTTL or DefaultTTL is more than 31536000 seconds, then the default
            value for this field is the same as the value of DefaultTTL.

        min_ttl (int):
            The minimum amount of time, in seconds, that you want objects to stay in the CloudFront cache
            before CloudFront sends another request to the origin to see if the object has been updated. For
            more information, see Managing How Long Content Stays in an Edge Cache (Expiration) in the
            Amazon CloudFront Developer Guide.

        parameter_in_cache_key_and_forward_to_origin (Dict[str, Any], optional):
            The HTTP headers, cookies, and URL query strings to include in the cache key. The values
            included in the cache key are automatically included in requests that CloudFront sends to the
            origin.

            * EnableAcceptEncodingGzip (bool):
                A flag that can affect whether the Accept-Encoding HTTP header is
                included in the cache key and included in requests that CloudFront sends to the origin. This field is
                related to the EnableAcceptEncodingBrotli field. If one or both of these fields is true and the viewer
                request includes the Accept-Encoding header, then CloudFront does the following:
                Normalizes the value of the viewer’s Accept-Encoding header   Includes the normalized header in the
                cache key Includes the normalized header in the request to the origin, if a request is necessary.
                For more information, see Compression support in the Amazon CloudFront Developer Guide. If you set
                this value to true, and this cache behavior also has an origin request policy attached, do not
                include the Accept-Encoding header in the origin request policy. CloudFront always includes the
                Accept-Encoding header in origin requests when the value of this field is true, so including
                this header in an origin request policy has no effect. If both of these fields are false, then
                CloudFront treats the Accept-Encoding header the same as any other HTTP header in the viewer
                request. By default, it’s not included in the cache key and it’s not included in origin
                requests. In this case, you can manually add Accept-Encoding to the headers whitelist like any
                other HTTP header.

            * EnableAcceptEncodingBrotli (bool, optional):
                A flag that can affect whether the Accept-Encoding HTTP header is included in the cache key and
                included in requests that CloudFront sends to the origin. This field is related to the
                EnableAcceptEncodingGzip field. If one or both of these fields is true and the viewer request
                includes the Accept-Encoding header, then CloudFront does the following:   Normalizes the value
                of the viewer’s Accept-Encoding header   Includes the normalized header in the cache key
                Includes the normalized header in the request to the origin, if a request is necessary   For
                more information, see Compression support in the Amazon CloudFront Developer Guide. If you set
                this value to true, and this cache behavior also has an origin request policy attached, do not
                include the Accept-Encoding header in the origin request policy. CloudFront always includes the
                Accept-Encoding header in origin requests when the value of this field is true, so including
                this header in an origin request policy has no effect. If both of these fields are false, then
                CloudFront treats the Accept-Encoding header the same as any other HTTP header in the viewer
                request. By default, it’s not included in the cache key and it’s not included in origin
                requests. In this case, you can manually add Accept-Encoding to the headers whitelist like any
                other HTTP header.

            * HeadersConfig (Dict[str, Any]):
                An object that determines whether any HTTP headers (and if so, which headers) are included in
                the cache key and automatically included in requests that CloudFront sends to the origin.

                * HeaderBehavior (str):
                    Determines whether any HTTP headers are included in the cache key and automatically
                    included in requests that CloudFront sends to the origin. Valid values are:    none – HTTP headers are not
                    included in the cache key and are not automatically included in requests that CloudFront sends to the
                    origin. Even when this field is set to none, any headers that are listed in an OriginRequestPolicy
                    are included in origin requests.    whitelist – The HTTP headers that are listed in the Headers type
                    are included in the cache key and are automatically included in requests that CloudFront sends to the
                    origin.

                    * Headers (Dict[str, Any], optional): Contains a list of HTTP header names.
                        * Quantity (int): The number of header names in the Items list.
                        * Items (list[str], optional): A list of HTTP header names.

            * CookiesConfig (Dict[str, Any]):
                An object that determines whether any cookies in viewer requests (and if so, which cookies) are
                included in the cache key and automatically included in requests that CloudFront sends to the
                origin.

                * CookieBehavior (str):
                    Determines whether any cookies in viewer requests are included in the cache key and
                    automatically included in requests that CloudFront sends to the origin. Valid values are:
                    none – Cookies in viewer requests are not included in the cache key and are not automatically
                    included in requests that CloudFront sends to the origin. Even when this field is set to none,
                    any cookies that are listed in an OriginRequestPolicy are included in origin requests.
                    whitelist – The cookies in viewer requests that are listed in the CookieNames type are included
                    in the cache key and automatically included in requests that CloudFront sends to the origin.
                    allExcept – All cookies in viewer requests that are  not  listed in the CookieNames type are
                    included in the cache key and automatically included in requests that CloudFront sends to the
                    origin.    all – All cookies in viewer requests are included in the cache key and are
                    automatically included in requests that CloudFront sends to the origin.

                    * Cookies (Dict[str, Any], optional):
                        Contains a list of cookie names.
                        * Quantity (int): The number of cookie names in the Items list.
                        * Items (list[str], optional): A list of cookie names.

            * QueryStringsConfig (Dict[str, Any]):
                An object that determines whether any URL query strings in viewer requests (and if so, which
                query strings) are included in the cache key and automatically included in requests that
                CloudFront sends to the origin.

                * QueryStringBehavior (str):
                    Determines whether any URL query strings in viewer requests are included in the cache key and
                    automatically included in requests that CloudFront sends to the origin. Valid values are:
                    none – Query strings in viewer requests are not included in the cache key and are not
                    automatically included in requests that CloudFront sends to the origin. Even when this field is
                    set to none, any query strings that are listed in an OriginRequestPolicy are included in origin
                    requests.    whitelist – The query strings in viewer requests that are listed in the
                    QueryStringNames type are included in the cache key and automatically included in requests that
                    CloudFront sends to the origin.    allExcept – All query strings in viewer requests that are
                    not  listed in the QueryStringNames type are included in the cache key and automatically
                    included in requests that CloudFront sends to the origin.    all – All query strings in viewer
                    requests are included in the cache key and are automatically included in requests that
                    CloudFront sends to the origin.

                    * QueryStrings (Dict[str, Any], optional):
                        Contains the specific query strings in viewer requests that either  are  or  are not  included
                        in the cache key and automatically included in requests that CloudFront sends to the origin. The
                        behavior depends on whether the QueryStringBehavior field in the CachePolicyQueryStringsConfig
                        type is set to whitelist (the listed query strings  are  included) or allExcept (the listed
                        query strings  are not  included, but all other query strings are).

                        * Quantity (int): The number of query string names in the Items list.
                        * Items (list[str], optional): A list of query string names.

    Request Syntax:

        .. code-block:: sls

            resource_is_present:
              aws_auto.cloudfront.cache_policy.present:
                - name: "string"
                - comment: "string"
                - default_ttl: "int"
                - max_ttl: "int"
                - min_ttl: "int"
                - parameter_in_cache_key_and_forward_to_origin:
                    CookiesConfig:
                      CookieBehavior: "string"
                    EnableAcceptEncodingBrotli: "bool"
                    EnableAcceptEncodingGzip: "bool"
                    HeadersConfig:
                      HeaderBehavior: "string"
                    QueryStringsConfig:
                      QueryStringBehavior: "string"
                - etag: "string"

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              aws_auto.cloudfront.cache_policy.present:
                - name: value
                - comment: idem test
                - default_ttl: 86400
                - max_ttl: 31536000
                - min_ttl: 1
                - parameter_in_cache_key_and_forward_to_origin:
                    CookiesConfig:
                      CookieBehavior: none
                    EnableAcceptEncodingBrotli: true
                    EnableAcceptEncodingGzip: true
                    HeadersConfig:
                      HeaderBehavior: none
                    QueryStringsConfig:
                      QueryStringBehavior: none
                - etag: E23ZP02F085DFQ
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.aws.cloudfront.cache_policy.get(
            ctx, name=name, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result
        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])

        # check update
        update_ret = await hub.tool.aws.cloudfront.cloudfront_utils.update_cache_policy(
            ctx=ctx,
            name=name,
            before=result["old_state"],
            resource_id=resource_id,
            comment=comment,
            default_ttl=default_ttl,
            max_ttl=max_ttl,
            min_ttl=min_ttl,
            parameter_in_cache_key_and_forward_to_origin=parameter_in_cache_key_and_forward_to_origin,
            if_match=before["ret"].get("etag"),
        )
        result["comment"] = update_ret["comment"]
        result["result"] = update_ret["result"]
        resource_updated = bool(update_ret["ret"])
        if update_ret["ret"]:
            if ctx.get("test", False):
                for modified_param in update_ret["ret"]:
                    plan_state[modified_param] = update_ret["ret"][modified_param]
                result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.cloudfront.cache_policy", name=name
                )
            else:
                result["comment"] += hub.tool.aws.comment_utils.update_comment(
                    resource_type="aws.cloudfront.cache_policy", name=name
                )

    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "comment": comment,
                    "default_ttl": default_ttl,
                    "max_ttl": max_ttl,
                    "min_ttl": min_ttl,
                    "parameter_in_cache_key_and_forward_to_origin": parameter_in_cache_key_and_forward_to_origin,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.cloudfront.cache_policy", name=name
            )
            return result

        cache_policy_config = {
            "Comment": comment,
            "Name": name,
            "DefaultTTL": default_ttl,
            "MaxTTL": max_ttl,
            "MinTTL": min_ttl,
            "ParametersInCacheKeyAndForwardedToOrigin": parameter_in_cache_key_and_forward_to_origin,
        }
        cache_policy_config = (
            hub.tool.aws.cloudfront.cloudfront_utils.sanitize_policy_config(
                cache_policy_config
            )
        )
        ret = await hub.exec.boto3.client.cloudfront.create_cache_policy(
            ctx, CachePolicyConfig=cache_policy_config
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.cloudfront.cache_policy", name=name
        )
        resource_id = ret["ret"]["CachePolicy"]["Id"]

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif (not before) or resource_updated:
        after = await hub.exec.aws.cloudfront.cache_policy.get(
            ctx, name=name, resource_id=resource_id
        )
        if not after["result"]:
            result["result"] = False
            result["comment"] = after["comment"]
            return result
        result["new_state"] = copy.deepcopy(after["ret"])
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """
    Deletes a cache policy.

    You cannot delete a cache policy if it’s attached to a cache behavior. First update your distributions to remove the
    cache policy from all cache behaviors, then delete the cache policy. To delete a cache policy, you must provide the
    policy’s identifier and version.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            An identifier of the resource in the provider.

    Request Syntax:

        .. code-block:: sls

            resource_is_absent:
              aws_auto.cloudfront.cache_policy.absent:
                - name: "string"
                - resource_id: "string"

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws_auto.cloudfront.cache_policy.absent:
                - name: value
                - resource_id: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.cloudfront.cache_policy", name=name
        )
        return result

    before = await hub.exec.aws.cloudfront.cache_policy.get(
        ctx, name=name, resource_id=resource_id
    )

    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result

    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.cloudfront.cache_policy", name=name
        )

    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.cloudfront.cache_policy", name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]

        ret = await hub.exec.boto3.client.cloudfront.delete_cache_policy(
            ctx, Id=resource_id, IfMatch=before["ret"]["etag"]
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result

        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.cloudfront.cache_policy", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws_auto.cloudfront.cache_policy
    """
    result = {}

    ret = await hub.exec.aws.cloudfront.cache_policy.list(ctx)
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    for policy in ret["ret"]:
        resource_id = policy.get("resource_id")

        result[resource_id] = {
            "aws.cloudfront.cache_policy.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in policy.items()
            ]
        }

    return result



"""Util functions for SQS Queue."""
from typing import Any
from typing import Dict


DEFAULT_ATTRIBUTES_VALUES = {
    "delay_seconds": 0,
    "maximum_message_size": 262144,
    "message_retention_period": 345600,
    "policy": None,
    "receive_message_wait_time_seconds": 0,
    "redrive_policy": None,
    "visibility_timeout": 30,
    "kms_master_key_id": None,
    "kms_data_key_reuse_period_seconds": 300,
    "sqs_managed_sse_enabled": False,
    "fifo_queue": False,
    "content_based_deduplication": False,
    "deduplication_scope": "queue",
    "fifo_throughput_limit": "perQueue",
}


def compare_present_queue_attributes(
    hub, expected_attributes: Dict[str, Any], actual_attributes: Dict[str, Any]
) -> bool:
    """Checks if the expected_attributes are contained withing actual_attributes.

    A None value in expected_attributes and a default value in actual_attributes are considered equal.

    Args:
        expected_attributes(Dict):
            The expected attributes

        actual_attributes(Dict):
            The actual attributes

    Returns:
        True if expected_attributes is contained within actual_attributes, False otherwise
    """
    for (
        expected_attribute_name,
        expected_attribute_value,
    ) in expected_attributes.items():
        actual_attribute_value = actual_attributes.get(expected_attribute_name)

        if actual_attribute_value is None:
            return actual_attribute_value == expected_attribute_value

        if expected_attribute_name == "policy":
            if not hub.tool.aws.state_comparison_utils.are_policies_equal(
                expected_attribute_value, actual_attribute_value
            ):
                return False
            continue

        if expected_attribute_value != actual_attribute_value:
            if (
                expected_attribute_value is None
                and actual_attributes.get(expected_attribute_name)
                == DEFAULT_ATTRIBUTES_VALUES[expected_attribute_name]
            ):
                continue

            return False

    return True



"""Exec module for managing Guardduty org admins."""
__func_alias__ = {"list_": "list"}

from typing import Dict


async def list_(hub, ctx, name: str = None) -> Dict:
    """List AWS Guardduty organization admin accounts.

    Fetch a list of organization admin accounts for AWS Guardduty. The function returns empty list when no resource is found.
    Only the organization's management account can run this API operation.

    Args:
        name(str, Optional):
            The name of the Idem state.

    Returns:
        Dict[str, Any]:
            Returns organization admin accounts in present format

    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec aws.guardduty.organization_admin_account.list name="list org admins"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resources:
              exec.run:
                - path: aws.guardduty.organization_admin_account.list
                - kwargs:
                    name: my_resources
    """
    result = dict(comment=[], ret=[], result=True)
    ret = await hub.exec.boto3.client.guardduty.list_organization_admin_accounts(ctx)
    if not ret["result"]:
        result["result"] = ret["result"]
        result["comment"] += list(ret["comment"])
        return result
    if not ret["ret"]["AdminAccounts"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.guardduty.organization_admin_account", name=name
            )
        )
        return result
    for admin_account in ret["ret"]["AdminAccounts"]:
        result["ret"].append(
            hub.tool.aws.guardduty.conversion_utils.convert_raw_org_admin_account_to_present(
                raw_resource=admin_account, idem_resource_name=name
            )
        )
    return result



"""Exec module for managing EC2 ami."""
from datetime import datetime
from typing import Dict
from typing import List

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    executable_users: List = None,
    owners: List = None,
    include_deprecated: bool = False,
    most_recent: bool = True,
    filters: List = None,
    tags: Dict[str, str] = None,
) -> Dict:
    """Get EC2 ami from AWS account.

    Get a single ami from AWS. Set most_recent=True to return most recent image if more than one resource is found.
    Otherwise, the first resource returned from AWS will be used. The function returns None when no resource is found.

    Args:
        name (str):
            The name of the Idem state.

        resource_id (str, Optional):
            Image id of the ami.

        executable_users (list[str], Optional):
            Scopes the images by users with explicit launch permissions.
            Specify an Amazon Web Services account ID, self (the sender of the request), or all (public AMIs).

        owners (list[str], Optional):
            Scopes the results to images with the specified owners.
            You can specify a combination of Amazon Web Services account IDs, self , amazon , and aws-marketplace.
            If you omit this parameter, the results include all images for which you have launch permissions, regardless of ownership.

        include_deprecated (bool, Optional):
            If true , all deprecated AMIs are included in the response.
            If false , no deprecated AMIs are included in the response. If no value is specified, the default value is false .

        most_recent (bool, Optional):
            If more than one result is returned, use the most recent AMI.

        filters (list[str, str], Optional):
            One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_images

        tags (dict[str, str], Optional):
            The dict of tags to filter by in the format ``{tag-key: tag-value}`` . For example, to find all resources
            that have a tag with the key "Owner" and the value "TeamA" , specify ``{Owner: TeamA}``

            * Key (str):
                The key name for the tag to be used to filter by.

            * Value(str):
                The value associated with this tag to filter by.

    Returns:
        Dict[str, Any]:
            Returns ami in present format

    Examples:
        Calling this exec module function from the cli with filters

        .. code-block:: bash

            idem exec aws.ec2.ami.get name="my_resource" most_recent="True"  filters=[{'name': 'name', 'values': ['ami-name']}, {'name': 'virtualization-type', 'values': ['hvm']}]

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.ec2.ami.get
                - kwargs:
                    name: my_resource
                    most_recent: True
                    filters:
                        - name: 'name'
                          values: ["ami-name"]
                        - name: 'virtualization-type'
                          values: ["hvm"]
    """
    result = dict(comment=[], ret=None, result=True)

    tags = hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)

    ret = await hub.tool.aws.ec2.ami.search_raw(
        ctx=ctx,
        resource_id=resource_id,
        filters=filters,
        executable_users=executable_users,
        owners=owners,
        include_deprecated=include_deprecated,
        tags=tags,
    )

    if not ret["result"]:
        if "InvalidAMIID.Malformed" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.ec2.ami", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["Images"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ec2.ami", name=name
            )
        )
        return result

    resource = ret["ret"]["Images"][0]
    if len(ret["ret"]["Images"]) > 1:
        if most_recent:
            ret["ret"]["Images"].sort(
                key=lambda x: datetime.strptime(
                    x.get("CreationDate"), "%Y-%m-%dT%H:%M:%S.%fZ"
                ),
                reverse=True,
            )
            resource = ret["ret"]["Images"][0]
        else:
            result["comment"].append(
                hub.tool.aws.comment_utils.find_more_than_one(
                    resource_type="aws.ec2.ami", resource_id=resource_id
                )
            )

    result["ret"] = hub.tool.aws.ec2.conversion_utils.convert_raw_ami_to_present(
        raw_resource=resource
    )

    return result


async def list_(
    hub,
    ctx,
    name: str = None,
    executable_users: List = None,
    owners: List = None,
    include_deprecated: bool = False,
    filters: List = None,
    tags: Dict[str, str] = None,
) -> Dict:
    """List EC2 ami from AWS account.

    Fetch a list of ami from AWS. The function returns empty list when no resource is found.

    Args:
        name(str, Optional):
            The name of the Idem state.

        executable_users(list[str], Optional):
            Scopes the images by users with explicit launch permissions.
            Specify an Amazon Web Services account ID, self (the sender of the request), or all (public AMIs).

        owners(list[str], Optional):
            Scopes the results to images with the specified owners.
            You can specify a combination of Amazon Web Services account IDs, self , amazon , and aws-marketplace.
            If no value is specified, the default value is ["self"].

        include_deprecated(bool, Optional):
            If true, all deprecated AMIs are included in the response.
            If false, no deprecated AMIs are included in the response. If no value is specified, the default value is false.

        filters(list, Optional):
            One or more filters, A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_images

        tags(dict, Optional):
            The dict of tags to filter by in the format ``{tag-key: tag-value}`` . For example, to find all resources
            that have a tag with the key "Owner" and the value "TeamA" , specify ``{Owner: TeamA}``

            * Key(str):
                The key name for the tag to be used to filter by.

            * Value(str):
                The value associated with this tag to filter by.

    Returns:
        Dict[str, Any]:
            Returns ami in present format

    Examples:
        Calling this exec module function from the cli with filters

        .. code-block:: bash

            idem exec aws.ec2.ami.list filters=[{'name': 'name', 'values': ['ami-name']}, {'name': 'virtualization-type', 'values': ['hvm']}]

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.ec2.ami.list
                - kwargs:
                    filters:
                        - name: 'name'
                          values: ["ami-name"]
                        - name: 'virtualization-type'
                          values: ["hvm"]
    """
    result = dict(comment=[], ret=[], result=True)

    tags = hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)

    ret = await hub.tool.aws.ec2.ami.search_raw(
        ctx=ctx,
        filters=filters,
        executable_users=executable_users,
        owners=owners,
        include_deprecated=include_deprecated,
        tags=tags,
    )

    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["Images"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.ec2.ami", name=name
            )
        )
        return result

    for image in ret["ret"]["Images"]:
        result["ret"].append(
            hub.tool.aws.ec2.conversion_utils.convert_raw_ami_to_present(
                raw_resource=image
            )
        )
    return result



"""State module for managing Amazon Cloudformation Stacks."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    template_body: str = None,
    template_url: str = None,
    parameters: List[
        make_dataclass(
            "Parameters",
            [
                ("KeyName", dict, field(default=None)),
                ("SSHLocation", dict, field(default=None)),
                ("InstanceType", dict, field(default=None)),
            ],
        )
    ] = None,
    disable_rollback: bool = None,
    rollback_configuration: make_dataclass(
        "RollbackConfiguration",
        [
            (
                "RollbackTriggers",
                make_dataclass(
                    "RollbackTriggers",
                    [
                        ("Arn", str, field(default=None)),
                        ("Type", str, field(default=None)),
                    ],
                ),
            ),
            ("MonitoringTimeInMinutes", str),
        ],
    ) = None,
    timeout: make_dataclass(
        "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=None)),
                        ("max_attempts", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=None)),
                        ("max_attempts", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
    notification_arns: List = None,
    capabilities: List = None,
    resource_types: List = None,
    role_arn: str = None,
    on_failure: str = None,
    stack_policy_body: str = None,
    stack_policy_url: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
    client_request_token: str = None,
    enable_termination_protection: bool = None,
) -> Dict[str, Any]:
    """Creates a Stack as specified in the template. After the call completes successfully, the stack creation starts.

    Args:
        name(str):
            The name of the Cloudformation Stack. The name must be unique in the Region in which you are creating the
            stack. A stack name can contain only alphanumeric characters (case sensitive) and hyphens. It must start
            with an alphabetical character and can't be longer than 128 characters.

        resource_id(str, Optional):
            The ID that's associated with the stack.

        template_body(str, Optional):
            Structure containing the template body with a minimum length of 1 byte and a maximum length of 51,200 bytes.
            For more information, go to Template anatomy in the CloudFormation User Guide. Conditional: You must specify
            either the TemplateBody or the TemplateURL parameter, but not both. Defaults to None.

        template_url(str, Optional):
            Location of file containing the template body. The URL must point to a template (max size: 460,800 bytes)
            that's located in an Amazon S3 bucket or a Systems Manager document. For more information, go to the Template
            anatomy in the CloudFormation User Guide. Conditional: You must specify either the TemplateBody or the
            TemplateURL parameter, but not both. Defaults to None.

        parameters(list, Optional):
            A list of Parameter structures that specify input parameters for the stack. For more information, see the
            Parameter data type. Supported parameter keys are KeyName, InstanceType, and SSHLocation, each with their
            respective dict type keys. Defaults to None.

        disable_rollback(bool, Optional):
            Set to true to disable rollback of the stack if stack creation failed. You can specify either DisableRollback
            or OnFailure, but not both. Default: false. Defaults to None.

        rollback_configuration(dict, Optional):
            The rollback triggers for CloudFormation to monitor during stack creation and updating operations, and for
            the specified monitoring period afterwards. Its key, RollbackTriggers(list), triggers to monitor during stack
            creation or update actions. By default, CloudFormation saves the rollback triggers specified for a stack and
            applies them to any subsequent update operations for the stack, unless you specify otherwise. If you do specify
            rollback triggers for this parameter, those triggers replace any list of triggers previously specified for
            the stack. This means: (1) To use the rollback triggers previously specified for this stack, if any, don't
            specify this parameter. (2) To specify new or updated rollback triggers, you must specify all the triggers
            that you want used for this stack, even triggers you've specified before (for example, when creating the stack
            or during a previous stack update). Any triggers that you don't include in the updated list of triggers are
            no longer applied to the stack. (3) To remove all currently specified triggers, specify an empty list for
            this parameter. If a specified trigger is missing, the entire stack operation fails and is rolled back. A
            Rollback Trigger(dict) CloudFormation monitors during creation and updating of stacks. If any of the alarms
            you specify goes to ALARM state during the stack operation or within the specified monitoring period afterwards,
            CloudFormation rolls back the entire stack operation. The key for the Rollback Trigger, Arn (str), is the
            Amazon Resource Name (ARN) of the rollback trigger. If a specified trigger is missing, the entire stack
            operation fails and is rolled back. The value of the Rollback Trigger, Type (str), is resource type of the
            rollback trigger. Specify either AWS::CloudWatch::Alarm or AWS::CloudWatch::CompositeAlarm resource types.
            Defaults to None.

        timeout(dict, Optional):
            Timeout configuration for create/update of AWS Cloudformation Stack.

            * create (dict):
                Timeout configuration for creating AWS Cloudformation Stack.

                * delay (int, Optional): The amount of time in seconds to wait between attempts.
                * max_attempts (int, Optional): Customized timeout configuration containing delay and max attempts.

            * update(dict, Optional):
                Timeout configuration for updating AWS Cloudformation Stack.

                * delay (int, Optional): The amount of time in seconds to wait between attempts.

                * max_attempts: (int, Optional) Customized timeout configuration containing delay and max attempts.


        notification_arns(list, Optional):
            The Amazon Simple Notification Service (Amazon SNS) topic ARNs to publish stack related events. You can find
            your Amazon SNS topic ARNs using the Amazon SNS console or your Command Line Interface (CLI). Defaults to None.

        capabilities(list, Optional):
            In some cases, you must explicitly acknowledge that your stack template contains certain capabilities in
            order for CloudFormation to create the stack. CAPABILITY_IAM and CAPABILITY_NAMED_IAM  Some stack templates
            might include resources that can affect permissions in your Amazon Web Services account; for example, by
            creating new Identity and Access Management (IAM) users. For those stacks, you must explicitly acknowledge
            this by specifying one of these capabilities. The following IAM resources require you to specify either
            the CAPABILITY_IAM or CAPABILITY_NAMED_IAM capability. If you have IAM resources, you can specify either
            capability. If you have IAM resources with custom names, you must specify CAPABILITY_NAMED_IAM. If you
            don't specify either of these capabilities, CloudFormation returns an InsufficientCapabilities error. If
            your stack template contains these resources, we recommend that you review all permissions associated with
            them and edit their permissions if necessary. AWS::IAM::AccessKey AWS::IAM::Group AWS::IAM::InstanceProfile
            AWS::IAM::Policy AWS::IAM::Role AWS::IAM::User AWS::IAM::UserToGroupAddition For more information, see
            Acknowledging IAM Resources in CloudFormation Templates. CAPABILITY_AUTO_EXPAND Some templates contain
            macros. Macros perform custom processing on templates; this can include simple actions like find-and-replace
            operations, all the way to extensive transformations of entire templates. Because of this, users typically
            create a change set from the processed template, so that they can review the changes resulting from the
            macros before actually creating the stack. If your stack template contains one or more macros, and you choose
            to create a stack directly from the processed template, without first reviewing the resulting changes in a
            change set, you must acknowledge this capability. This includes the AWS::Include and AWS::Serverless transforms,
            which are macros hosted by CloudFormation. If you want to create a stack from a stack template that contains
            macros and nested stacks, you must create the stack directly from the template using this capability. You
            should only create stacks directly from a stack template that contains macros if you know what processing the
            macro performs. Each macro relies on an underlying Lambda service function for processing stack templates.
            Be aware that the Lambda function owner can update the function operation without CloudFormation being notified.
            For more information, see Using CloudFormation macros to perform custom processing on templates. Defaults to
            None.

        resource_types(list, Optional):
            The template resource types that you have permissions to work with for this create stack action, such as
            AWS::EC2::Instance, AWS::EC2::*, or Custom::MyCustomInstance. Use the following syntax to describe template
            resource types: AWS::* (for all Amazon Web Services resources), Custom::* (for all custom resources),
            Custom::logical_ID  (for a specific custom resource), AWS::service_name::* (for all resources of a particular
            Amazon Web Services service), and AWS::service_name::resource_logical_ID  (for a specific Amazon Web Services
            resource). If the list of resource types doesn't include a resource that you're creating, the stack creation
            fails. By default, CloudFormation grants permissions to all resource types. Identity and Access Management
            (IAM) uses this parameter for CloudFormation-specific condition keys in IAM policies. For more information,
            see Controlling Access with Identity and Access Management. Defaults to None.

        role_arn(str, Optional):
            The Amazon Resource Name (ARN) of an Identity and Access Management (IAM) role that CloudFormation assumes
            to create the stack. CloudFormation uses the role's credentials to make calls on your behalf. CloudFormation
            always uses this role for all future operations on the stack. Provided that users have permission to operate
            on the stack, CloudFormation uses this role even if the users don't have permission to pass it. Ensure
            that the role grants least privilege. If you don't specify a value, CloudFormation uses the role that was
            previously associated with the stack. If no role is available, CloudFormation uses a temporary session that's
            generated from your user credentials. Defaults to None.

        on_failure(str, Optional):
            Determines what action will be taken if stack creation fails. This must be one of: DO_NOTHING, ROLLBACK, or
            DELETE. You can specify either OnFailure or DisableRollback, but not both. Default: ROLLBACK. Defaults to None.

        stack_policy_body(str, Optional):
            Structure containing the stack policy body. For more information, go to  Prevent Updates to Stack Resources
            in the CloudFormation User Guide. You can specify either the StackPolicyBody or the StackPolicyURL parameter,
            but not both. Defaults to None.

        stack_policy_url(str, Optional):
            Location of a file containing the stack policy. The URL must point to a policy (maximum size: 16 KB) located
            in an S3 bucket in the same Region as the stack. You can specify either the StackPolicyBody or the
            StackPolicyURL parameter, but not both. Defaults to None.

        tags(list, Optional):
            Key-value pairs to associate with this stack. CloudFormation also propagates these tags to the resources
            created in the stack. A maximum number of 50 tags can be specified. Defaults to None.

        client_request_token(str, Optional):
            A unique identifier for this CreateStack request. Specify this token if you plan to retry requests so that
            CloudFormation knows that you're not attempting to create a stack with the same name. You might retry
            CreateStack requests to ensure that CloudFormation successfully received them. All events initiated by a
            given stack operation are assigned the same client request token, which you can use to track operations.
            For example, if you execute a CreateStack operation with the token token1, then all the StackEvents generated
            by that operation will have ClientRequestToken set as token1. In the console, stack operations display the
            client request token on the Events tab. Stack operations that are initiated from the console use the token
            format Console-StackOperation-ID, which helps you easily identify the stack operation. For example, if you
            create a stack using the console, each stack event would be assigned the same token in the following format:
            Console-CreateStack-7f59c3cf-00d2-40c7-b2ff-e75db0987002. Defaults to None.

        enable_termination_protection(bool, Optional):
            Whether to enable termination protection on the specified stack. If a user attempts to delete a stack with
            termination protection enabled, the operation fails and the stack remains unchanged. For more information,
            see Protecting a Stack From Being Deleted in the CloudFormation User Guide. Termination protection is
            deactivated on stacks by default. For nested stacks, termination protection is set on the root stack and
            can't be changed directly on the nested stack. Defaults to None.

    Returns:
        dict[str, Any]

    Request Syntax:
        .. code-block:: sls

            [idem_test_aws_cloudformation_stack]:
              aws.cloudformation.stack.present:
                - name: string
                - template_body: string OR template_url: string
                - parameters: dict
                - disable_rollback: boolean
                - rollback_configuration: dict
                - timeout: dict
                - notification_arns: list
                - capabilities: list
                - resource_types: list
                - role_arn: string
                - on_failure: string
                - stack_policy_body: string
                - stack_policy_url: string
                - tags: list
                - client_request_token: string
                - enable_termination_protection: boolean

    Examples:
        .. code-block:: sls

            idem_test_aws_cloudformation_stack:
              aws.cloudformation.stack.present:
                - name: 'idem_test_cloudformation_stack'
                - template_body: "https://s3.amazonaws.com/cloudformation-templates-us-east-1/EC2WithSecurityGroup.template"

    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False

    if resource_id:
        before = await hub.exec.aws.cloudformation.stack.get(
            ctx, name=name, resource_id=resource_id
        )

        if not before["result"] or not before["ret"]:
            result["comment"] = before["comment"]
            result["result"] = False
            return result

        result["old_state"] = before["ret"]
        result["new_state"] = copy.deepcopy(result["old_state"])

        update_ret = await hub.tool.aws.cloudformation.stack.update(
            ctx,
            stack_name=name,
            old_state=result["old_state"],
            template_body=template_body,
            template_url=template_url,
            parameters=parameters,
            capabilities=capabilities,
            resource_types=resource_types,
            role_arn=role_arn,
            rollback_configuration=rollback_configuration,
            notification_arns=notification_arns,
            tags=tags,
            client_request_token=client_request_token,
            stack_policy_body=stack_policy_body,
            stack_policy_url=stack_policy_url,
            disable_rollback=disable_rollback,
        )
        result["comment"] = result["comment"] + update_ret["comment"]

        if not update_ret["result"]:
            result["result"] = False
            return result
        resource_updated = bool(update_ret["ret"])

    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "template_url": template_url,
                    "template_body": template_body,
                    "parameters": parameters,
                    "capabilities": capabilities,
                    "resource_types": resource_types,
                    "role_arn": role_arn,
                    "rollback_configuration": rollback_configuration,
                    "notification_arns": notification_arns,
                    "tags": tags,
                    "client_request_token": client_request_token,
                    "stack_policy_body": stack_policy_body,
                    "stack_policy_url": stack_policy_url,
                    "disable_rollback": disable_rollback,
                    "timeout": timeout,
                    "on_failure": on_failure,
                    "enable_termination_protection": enable_termination_protection,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.cloudformation.stack", name=name
            )
            return result

        ret = await hub.exec.boto3.client.cloudformation.create_stack(
            ctx,
            StackName=name,
            TemplateBody=template_body,
            TemplateURL=template_url,
            Parameters=parameters,
            DisableRollback=disable_rollback,
            RollbackConfiguration=rollback_configuration,
            TimeoutInMinutes=timeout,
            NotificationARNs=notification_arns,
            Capabilities=capabilities,
            RoleARN=role_arn,
            OnFailure=on_failure,
            StackPolicyBody=stack_policy_body,
            StackPolicyURL=stack_policy_url,
            Tags=tags,
            ClientRequestToken=client_request_token,
            EnableTerminationProtection=enable_termination_protection,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.cloudformation.stack", name=name
        )
        resource_id = ret["ret"]["StackId"]

        create_waiter_acceptors = [
            {
                "argument": "Stacks[].StackStatus",
                "expected": "CREATE_COMPLETE",
                "matcher": "pathAll",
                "state": "success",
            },
            {
                "argument": "Stacks[].StackStatus",
                "expected": "CREATE_FAILED",
                "matcher": "pathAny",
                "state": "failure",
            },
            {
                "argument": "Stacks[].StackStatus",
                "expected": "DELETE_COMPLETE",
                "matcher": "pathAny",
                "state": "failure",
            },
            {
                "argument": "Stacks[].StackStatus",
                "expected": "DELETE_FAILED",
                "matcher": "pathAny",
                "state": "failure",
            },
            {
                "argument": "Stacks[].StackStatus",
                "expected": "ROLLBACK_FAILED",
                "matcher": "pathAny",
                "state": "failure",
            },
            {
                "argument": "Stacks[].StackStatus",
                "expected": "ROLLBACK_COMPLETE",
                "matcher": "pathAny",
                "state": "failure",
            },
            {"expected": "ValidationError", "matcher": "error", "state": "failure"},
        ]
        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=10,
            default_max_attempts=40,
            timeout_config=timeout.get("create") if timeout else None,
        )
        endpoint_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
            name="StackCreate",
            operation="DescribeStacks",
            argument=["Stacks[].StackStatus"],
            acceptors=create_waiter_acceptors,
            client=await hub.tool.boto3.client.get_client(ctx, "cloudformation"),
        )
        try:
            await hub.tool.boto3.client.wait(
                ctx,
                "cloudformation",
                "StackCreateComplete",
                endpoint_waiter,
                StackName=name,
                WaiterConfig=waiter_config,
            )
        except Exception as e:
            result["comment"] = str(e)
            result["result"] = False

    try:
        if ctx.get("test", False) and resource_updated:
            result["new_state"] = update_ret["ret"]
            return result

        elif (not before) or resource_updated:
            after = await hub.exec.aws.cloudformation.stack.get(
                ctx, name=name, resource_id=resource_id
            )
            if not (after["result"] and after["ret"]):
                result["result"] = False
                result["comment"] = result["comment"] + after["comment"]
                return result

            result["new_state"] = after["ret"]
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])

    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    role_arn: str = None,
    client_request_token: str = None,
    timeout: Dict = None,
    retain_resources: list = None,
) -> Dict[str, Any]:
    """
    Deletes a specified stack. Once the call completes successfully, stack deletion starts. Deleted stacks don't
    show up in the DescribeStacks operation if the deletion has been completed successfully.

    Args:
        name(str):
            The name of the Cloudformation Stack.

        resource_id(str, Optional):
            The unique stack ID that's associated with the stack. Idem automatically considers this resource
            being absent if this field is not specified.

        role_arn(str, Optional):
            The Amazon Resource Name (ARN) of an Identity and Access Management (IAM) role that CloudFormation assumes to
            delete the stack. CloudFormation uses the role's credentials to make calls on your behalf. If you don't specify
            a value, CloudFormation uses the role that was previously associated with the stack. If no role is available,
            CloudFormation uses a temporary session that's generated from your user credentials. Defaults to None.

        client_request_token(str, Optional):
            A unique identifier for this DeleteStack request. Specify this token if you plan to retry requests so that
            CloudFormation knows that you're not attempting to delete a stack with the same name. You might retry
            DeleteStack requests to ensure that CloudFormation successfully received them. All events initiated by a
            given stack operation are assigned the same client request token, which you can use to track operations.
            For example, if you execute a CreateStack operation with the token token1, then all the StackEvents generated
            by that operation will have ClientRequestToken set as token1. In the console, stack operations display the
            client request token on the Events tab. Stack operations that are initiated from the console use the token
            format Console-StackOperation-ID, which helps you easily identify the stack operation. For example, if you
            create a stack using the console, each stack event would be assigned the same token in the following format:
            Console-CreateStack-7f59c3cf-00d2-40c7-b2ff-e75db0987002. Defaults to None.

        timeout(dict, Optional):
            Timeout configuration for deletion of AWS Cloudformation Stack.

            * delete (str):
                Timeout configuration for deletion of a Cloudformation Stack.

                * delay -- The amount of time in seconds to wait between attempts.
                * max_attempts -- Customized timeout configuration containing delay and max attempts.

        retain_resources(list, Optional):
            For stacks in the DELETE_FAILED state, a list of resource logical IDs that are associated with the resources
            you want to retain. During deletion, CloudFormation deletes the stack but doesn't delete the retained
            resources. Retaining resources is useful when you can't delete a resource, such as a non-empty S3 bucket,
            but you want to delete the stack.


    Returns:
        dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.cloudformation.stack.absent:
                - name: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.cloudformation.stack", name=name
        )
        return result

    before_ret = await hub.exec.aws.cloudformation.stack.get(
        ctx, name=name, resource_id=resource_id
    )

    if not before_ret["result"]:
        result["result"] = False
        result["comment"] = before_ret["comment"]
        return result

    if not before_ret["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.cloudformation.stack", name=name
        )
        return result

    elif ctx.get("test", False):
        result["old_state"] = before_ret["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.cloudformation.stack", name=name
        )
        return result

    else:
        result["old_state"] = before_ret["ret"]

        ret = await hub.exec.boto3.client.cloudformation.delete_stack(
            ctx,
            StackName=name,
            RetainResources=retain_resources,
            RoleARN=role_arn,
            ClientRequestToken=client_request_token,
        )
        if not ret["result"]:
            result["result"] = False
            result["comment"] = ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.cloudformation.stack", name=name
        )

        delete_waiter_acceptors = [
            {
                "argument": "Stacks[].StackStatus",
                "expected": "DELETE_COMPLETE",
                "matcher": "pathAll",
                "state": "success",
            },
            {"expected": "ValidationError", "matcher": "error", "state": "success"},
            {
                "argument": "Stacks[].StackStatus",
                "expected": "DELETE_FAILED",
                "matcher": "pathAny",
                "state": "failure",
            },
            {
                "argument": "Stacks[].StackStatus",
                "expected": "CREATE_FAILED",
                "matcher": "pathAny",
                "state": "failure",
            },
            {
                "argument": "Stacks[].StackStatus",
                "expected": "ROLLBACK_FAILED",
                "matcher": "pathAny",
                "state": "failure",
            },
            {
                "argument": "Stacks[].StackStatus",
                "expected": "UPDATE_ROLLBACK_IN_PROGRESS",
                "matcher": "pathAny",
                "state": "failure",
            },
            {
                "argument": "Stacks[].StackStatus",
                "expected": "UPDATE_ROLLBACK_FAILED",
                "matcher": "pathAny",
                "state": "failure",
            },
            {
                "argument": "Stacks[].StackStatus",
                "expected": "UPDATE_ROLLBACK_COMPLETE",
                "matcher": "pathAny",
                "state": "failure",
            },
        ]
        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=10,
            default_max_attempts=40,
            timeout_config=timeout.get("delete") if timeout else None,
        )
        endpoint_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
            name="StackDelete",
            operation="DescribeStacks",
            argument=["Stacks[].StackStatus"],
            acceptors=delete_waiter_acceptors,
            client=await hub.tool.boto3.client.get_client(ctx, "cloudformation"),
        )
        try:
            await hub.tool.boto3.client.wait(
                ctx,
                "cloudformation",
                "StackDeleteComplete",
                endpoint_waiter,
                StackName=resource_id,
                WaiterConfig=waiter_config,
            )

        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False

        return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Describes the Cloudformation Stacks.

    Returns:
        dict[str, dict[str, Any]]

    Examples:

        .. code-block:: bash

            $ idem describe aws.cloudformation.stack
    """

    result = {}
    ret = await hub.exec.boto3.client.cloudformation.describe_stacks(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe stack {ret['comment']}")
        return result

    for stack in ret["ret"]["Stacks"]:
        resource_id = stack.get("StackId")
        resource_translated = (
            hub.tool.aws.cloudformation.stack.convert_raw_stack_to_present(
                raw_resource=stack,
                resource_id=resource_id,
                idem_resource_name=stack.get("StackName"),
            )
        )

        result[resource_translated["resource_id"]] = {
            "aws.cloudformation.stack.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result
